# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import glob
import json
import os
import platform
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple

from supervisor.datatypes import Automatic, logfile_name, boolean, integer, byte_size, logging_level, list_of_strings
from supervisor.loggers import Logger
from supervisor.options import (expand, make_namespec, ServerOptions,
                                ProcessConfig, FastCGIProcessConfig, EventListenerConfig)

from .ttypes import (ConciliationStrategies, EventLinks, StartingStrategies, SynchronizationOptions,
                     Ipv4Address, NameList, Payload, StatisticsTypes,
                     GroupConfigInfo, ProgramConfig, SupvisorsProcessConfig)


# Options of main section
def get_logger_configuration(**config) -> Payload:
    """ Extract the logger parameters from the config structure.

    Returns a dictionary with the following entries:
        - software_name: the optional software name, used as a prefix in log traces ;
        - logfile: absolute or relative path of the Supvisors log file ;
        - logfile_maxbytes: maximum size of the Supvisors log file ;
        - logfile_backups: number of Supvisors backup log files ;
        - loglevel: logging level.

    :param config: the configuration provided by Supervisor from the [rpcinterface:supvisors] section.
    :return: a dictionary containing the logger parameters.
    """
    return {'prefix': config.get('software_name', ''),
            'logfile': logfile_name(config.get('logfile', Automatic)),
            'logfile_maxbytes': byte_size(config.get('logfile_maxbytes', '50MB')),
            'logfile_backups': integer(config.get('logfile_backups', 10)),
            'loglevel': logging_level(config.get('loglevel', 'info'))}


class SupvisorsOptions:
    """ Holder of the Supvisors options.

    Attributes are:
        - software_name: a string to be displayed at the top of the Supvisors Web UI ;
        - software_icon: the path of an icon to be displayed at the top of the Supvisors Web UI ;
        - supvisors_list: list of Supvisors instance identifiers where Supvisors will be running ;
        - stereotype: the Supvisors instance stereotype, used as aliases in rules ;
        - multicast: UDP Multicast Group where Supvisors will exchange data ;
        - multicast_interface: UDP Multicast Group interface ;
        - multicast_ttl: UDP Multicast time-to-live ;
        - rules_files: list of absolute or relative paths to the XML rules files ;
        - event_link: type of the event link used to publish all Supvisors events ;
        - event_port: port number used to publish all Supvisors events ;
        - auto_fence: when True, Supvisors won't try to reconnect to a Supvisors instance that has been inactive ;
        - synchro_options: the conditions that will end the synchronization phase ;
        - synchro_timeout: time in seconds that Supvisors waits for all expected Supvisors instances to publish ;
        - inactivity_ticks: number of local ticks to wait before considering a remote Supvisors instance inactive ;
        - core_identifiers: subset of supvisors_list identifiers that will force the end of synchro when all RUNNING ;
        - disabilities_file: the file used to persist the process disabilities ;
        - conciliation_strategy: strategy used to solve conflicts when Supvisors has detected multiple running
          instances of the same program ;
        - starting_strategy: strategy used to start processes on Supvisors instances ;
        - host_stats_enabled: if False, no host statistics will be collected from this Supvisors instance ;
        - proc_stats_enabled: if False, no process statistics will be collected from this Supvisors instance ;
        - collecting_period: period of the statistics collection ;
        - stats_periods: list of periods for which the statistics will be provided in the Supvisors Web UI ;
        - stats_histo: depth of statistics history ;
        - stats_irix_mode: choice of CPU value display between IRIX and Solaris ;
        - tail_limit: the number of bytes used to display the log tail of the file in the Web UI (refresh mode) ;
        - tailf_limit: the number of bytes used to display the log tail of the file in the Web UI (tail -f mode).
    """

    SYNCHRO_TIMEOUT_MIN = 15
    SYNCHRO_TIMEOUT_MAX = 1200

    # default SynchronizationOptions list that is equivalent to previous Supvisors versions
    SYNCHRO_DEFAULT_OPTIONS = [SynchronizationOptions.STRICT,
                               SynchronizationOptions.TIMEOUT,
                               SynchronizationOptions.CORE]

    INACTIVITY_TICKS_MIN = 2
    INACTIVITY_TICKS_MAX = 720

    RESERVED_MULTICAST_ADDRESSES = ['224.0.0.0', '232.0.0.0', '233.0.0.0', '239.0.0.0']

    def __init__(self, supervisord, logger: Logger, **config):
        """ Initialization of the attributes.

        :param supervisord: the global Supervisor structure
        :param logger: the Supvisors logger
        :param config: the configuration provided by Supervisor from the [rpcinterface:supvisors] section
        """
        self.supervisord_options = supervisord.options
        self.logger = logger
        # get the string to be displayed at the top of the Supvisors Web UI
        self.software_name = self._get_value(config, 'software_name', '')
        self.software_icon = self._get_value(config, 'software_icon', None, self.to_existing_file)
        # get expected Supvisors instances
        self.supvisors_list = self._get_value(config, 'supvisors_list', None,
                                              lambda x: list(OrderedDict.fromkeys(filter(None, list_of_strings(x)))))
        # Supvisors instance generic type
        self.stereotypes = self._get_value(config, 'stereotypes', set(),
                                           lambda x: set(filter(None, list_of_strings(x))))
        # get multicast parameters for discovery mode
        self.multicast_group = self._get_value(config, 'multicast_group', None, self.to_multicast_group)
        self.multicast_interface = self._get_value(config, 'multicast_interface', None, self.to_ip_address)
        self.multicast_ttl = self._get_value(config, 'multicast_ttl', 1, self.to_ttl)
        # get the rules files
        self.rules_files = self._get_value(config, 'rules_files', None, self.to_filepaths)
        # if event_port is not defined, it will be set later based on Supervisor HTTP port
        self.event_link = self._get_value(config, 'event_link', EventLinks.NONE, self.to_event_link)
        self.event_port = self._get_value(config, 'event_port', 0, self.to_port_num)
        self.auto_fence = self._get_value(config, 'auto_fence', False, boolean)
        self.synchro_options = self._get_value(config, 'synchro_options', self.SYNCHRO_DEFAULT_OPTIONS,
                                               self.to_synchro_options)
        self.synchro_timeout = self._get_value(config, 'synchro_timeout', self.SYNCHRO_TIMEOUT_MIN, self.to_timeout)
        self.inactivity_ticks = self._get_value(config, 'inactivity_ticks', self.INACTIVITY_TICKS_MIN, self.to_ticks)
        # get the minimum list of identifiers to end the synchronization phase
        self.core_identifiers = self._get_value(config, 'core_identifiers', set(),
                                                lambda x: set(filter(None, list_of_strings(x))))
        # get disabilities file
        self.disabilities_file = self._get_value(config, 'disabilities_file', None, self.check_dirpath)
        # get strategies
        self.conciliation_strategy = self._get_value(config, 'conciliation_strategy', ConciliationStrategies.USER,
                                                     self.to_conciliation_strategy)
        self.starting_strategy = self._get_value(config, 'starting_strategy', StartingStrategies.CONFIG,
                                                 self.to_starting_strategy)
        # configure statistics
        # stats_enabled is deprecated
        stats_enabled = self._get_value(config, 'stats_enabled', (True, True), self.to_statistics_type)
        self.host_stats_enabled = stats_enabled[0]
        self.process_stats_enabled = stats_enabled[1]
        self.collecting_period = self._get_value(config, 'stats_collecting_period', 5, self.to_period)
        self.stats_periods = self._get_value(config, 'stats_periods', [10], self.to_periods)
        self.stats_histo = self._get_value(config, 'stats_histo', 200, self.to_histo)
        self.stats_irix_mode = self._get_value(config, 'stats_irix_mode', False, boolean)
        # configure log tail limits
        self.tail_limit = self._get_value(config, 'tail_limit', 1024, byte_size)
        self.tailf_limit = self._get_value(config, 'tailf_limit', 1024, byte_size)
        # check synchro options consistence
        self.check_synchro_options()

    def __str__(self):
        """ Contents as string. """
        mc_group = f'{self.multicast_group[0]}:{self.multicast_group[1]}' if self.multicast_group else None
        return (f'software_name="{self.software_name}"'
                f' software_icon={self.software_icon}'
                f' supvisors_list={self.supvisors_list}'
                f' stereotypes={self.stereotypes}'
                f' multicast_group={mc_group}'
                f' multicast_interface={self.multicast_interface}'
                f' multicast_ttl={self.multicast_ttl}'
                f' rules_files={self.rules_files}'
                f' event_link={self.event_link.name}'
                f' event_port={self.event_port}'
                f' auto_fence={self.auto_fence}'
                f' synchro_options={[x.name for x in self.synchro_options]}'
                f' synchro_timeout={self.synchro_timeout}'
                f' inactivity_ticks={self.inactivity_ticks}'
                f' core_identifiers={self.core_identifiers}'
                f' disabilities_file={self.disabilities_file}'
                f' conciliation_strategy={self.conciliation_strategy.name}'
                f' starting_strategy={self.starting_strategy.name}'
                f' host_stats_enabled={self.host_stats_enabled}'
                f' process_stats_enabled={self.process_stats_enabled}'
                f' collecting_period={self.collecting_period}'
                f' stats_periods={self.stats_periods}'
                f' stats_histo={self.stats_histo}'
                f' stats_irix_mode={self.stats_irix_mode}'
                f' tail_limit={self.tail_limit}'
                f' tailf_limit={self.tailf_limit}')

    @property
    def discovery_mode(self) -> bool:
        """ Return True if Supvisors is in discovery mode.

        :return: True if the multicast group is set
        """
        return self.multicast_group is not None

    def check_synchro_options(self):
        """ Check the validity of the synchro_options wrt other options. """
        # when using CORE in synchro_options, core_identifiers cannot be empty
        if not self.core_identifiers and SynchronizationOptions.CORE in self.synchro_options:
            self.logger.warn('SupvisorsOptions:check_synchro_options: cancellation of synchro_options CORE'
                             ' with no core_identifiers')
            self.synchro_options.remove(SynchronizationOptions.CORE)
        # when using LIST in synchro_options, supvisors_list cannot be empty
        if not self.supvisors_list and SynchronizationOptions.STRICT in self.synchro_options:
            self.logger.warn('SupvisorsOptions:check_synchro_options: cancellation of synchro_options STRICT'
                             ' with no supvisors_list')
            self.synchro_options.remove(SynchronizationOptions.STRICT)
        # finally, synchro_options must not be empty
        if not self.synchro_options:
            raise ValueError('synchro_options must not be empty')

    def check_dirpath(self, file_path: str) -> str:
        """ Check if the path provided exists and create the folder tree if necessary.
        update of Supervisor datatypes.existing_dirpath.

        :param file_path: the file path to check.
        :return: the file path.
        """
        expanded_file_path = os.path.expanduser(file_path)
        file_dir = os.path.dirname(expanded_file_path)
        if not file_dir:
            # relative pathname with no directory component
            return expanded_file_path
        if not os.path.isdir(file_dir):
            # if the folder path does not exist, try to create it
            try:
                self.logger.info(f'SupvisorsOptions.check_dirpath: creating folder={file_dir}')
                os.makedirs(file_dir)
            except PermissionError:
                # creation of the folder tree denied
                raise ValueError(f'The directory named as part of the path={file_path} cannot be created')
        return expanded_file_path

    def _get_value(self, config: Payload, attr: str, default_value, fct=None):
        """ Read and convert the option.

        :param config: the option dictionary.
        :param attr: the option considered.
        :param default_value: the default value to apply if not found in config or erroneous.
        :param fct: the optional conversion function to apply to the string value.
        :return: the option with the requested type.
        """
        if attr not in config:
            return default_value
        value = config[attr]
        if fct:
            try:
                return fct(value)
            except ValueError as exc:
                self.logger.error(f'SupvisorsOptions.get_value: {str(exc)}')
                return default_value
        return value

    # conversion utils (completion of supervisor.datatypes)
    def to_existing_file(self, value: str) -> Optional[str]:
        """ Expand the file globs and return the file found.
        One single result expected.

        :param value: a space-separated sequence of file globs.
        :return: the file found.
        """
        file_set = self.to_filepaths(value)
        candidates = [filepath for filepath in file_set
                      if os.path.isfile(filepath)]
        if len(candidates) > 1:
            self.logger.warn(f'SupvisorsOptions.to_existing_file: multiple candidates found matching {value}'
                             f' - {candidates}')
        return candidates[0] if candidates else None

    def to_filepaths(self, value: str) -> List[str]:
        """ Expand the file globs and return the files found.

        :param value: a space-separated sequence of file globs
        :return: the list of files found
        """
        # apply expansions to value
        expansions = {'here': self.supervisord_options.here,
                      'host_node_name': platform.node()}
        expansions.update(self.supervisord_options.environ_expansions)
        files = expand(value, expansions, 'rpcinterface.supvisors')
        # get all files
        file_set = set()
        for pattern in files.split():
            filepaths = glob.glob(pattern)
            for filepath in filepaths:
                file_set.add(os.path.abspath(filepath))
        # check that something came out
        if value and not file_set:
            self.logger.warn(f'SupvisorsOptions.to_filepaths: no file found matching {value}')
        return sorted(file_set)

    @staticmethod
    def to_multicast_group(value: str) -> Ipv4Address:
        """ Convert a string into a TTL number, in [0;255].

        :param value: the multicast address + port as a string
        :return: the verified multicast address + port
        """
        # parse the value to find address + port
        values = value.split(':', 1)
        if len(values) != 2:
            raise ValueError(f'invalid value for multicast_group: "{value}".'
                             f' "ip_address:port" expected')
        SupvisorsOptions._check_multicast_address(values[0])
        return values[0], SupvisorsOptions.to_port_num(values[1])

    @staticmethod
    def to_ip_address(value: str) -> Optional[str]:
        """ Check the formatting of the IP address.

        :param value: the IP address to check
        :return: None
        """
        # will set INADDR_ANY later
        if value in ['ANY', 'INADDR_ANY']:
            return None
        # parse the IP address
        try:
            values = value.split('.')
            if len(values) != 4:
                raise ValueError('wrong number of bytes')
            for idx in range(0, 4):
                SupvisorsOptions.to_integer(values[idx], f'IP byte {idx}', (0, 255))
        except ValueError:
            raise ValueError(f'invalid value for IP address: "{value}"')
        return value

    @staticmethod
    def _check_multicast_address(value: str):
        """ Check the formatting of the multicast address from 224.0.0.0 to 239.255.255.255.

        :param value: the multicast address to check
        :return: None
        """
        if value in SupvisorsOptions.RESERVED_MULTICAST_ADDRESSES:
            raise ValueError(f'reserved multicast address: "{value}".'
                             f' reserved addresses are {SupvisorsOptions.RESERVED_MULTICAST_ADDRESSES}')
        # parse the IP address
        try:
            values = value.split('.')
            if len(values) != 4:
                raise ValueError('wrong number of bytes')
            SupvisorsOptions.to_integer(values[0], 'multicast byte 1', (224, 239))
            for idx in range(1, 4):
                SupvisorsOptions.to_integer(values[idx], f'multicast byte {idx}', (0, 255))
        except ValueError:
            raise ValueError(f'invalid value for multicast address: "{value}".'
                             ' IP address expected from 224.0.0.0 to 239.255.255.255')

    @staticmethod
    def to_ttl(value: str) -> int:
        """ Convert a string into a TTL number, in [0;255].

        :param value: the TTL as a string
        :return: the TTL as an integer
        """
        return SupvisorsOptions.to_integer(value, 'multicast_ttl', (0, 255))

    @staticmethod
    def to_port_num(value: str) -> int:
        """ Convert a string into a port number, in [1;65535].

        :param value: the port number as a string
        :return: the port number as an integer
        """
        return SupvisorsOptions.to_integer(value, 'port', (1, 65535))

    @staticmethod
    def to_integer(value: str, type_name: str, limits: Tuple[int, int]) -> int:
        """ Convert a string into an integer within given limits.

        :param value: the integer as a string
        :param type_name: the integer nature for log in case of exception
        :param limits: the integer limits, given as a tuple(min, max) of inclusive bounds
        :return: the integer found
        """
        try:
            port = integer(value)
            if limits[0] > port or port > limits[1]:
                raise ValueError
            return port
        except ValueError:
            raise ValueError(f'invalid value for {type_name}: "{value}".'
                             f' integer expected in {limits}')

    @staticmethod
    def to_synchro_options(value: str) -> List[SynchronizationOptions]:
        """ Return the list of options selected to end the synchronization phase. """
        option_str_list = filter(None, list_of_strings(value))
        option_list = []
        for option_str in option_str_list:
            try:
                option = SynchronizationOptions[option_str.upper()]
            except KeyError:
                raise ValueError(f'invalid value for synchro_options: "{option_str}".'
                                 f' expected in {[x.name for x in SynchronizationOptions]})')
            if option not in option_list:
                option_list.append(option)
        return option_list

    @staticmethod
    def to_timeout(value: str) -> int:
        """ Convert a string into a timeout value, in [15;1200].

        :param value: the timeout as a string
        :return: the timeout as an integer
        """
        try:
            timeout = integer(value)
            if SupvisorsOptions.SYNCHRO_TIMEOUT_MIN > timeout or timeout > SupvisorsOptions.SYNCHRO_TIMEOUT_MAX:
                raise ValueError
            return timeout
        except ValueError:
            raise ValueError(f'invalid value for synchro_timeout: "{value}".'
                             f' integer expected in [{SupvisorsOptions.SYNCHRO_TIMEOUT_MIN};'
                             f'{SupvisorsOptions.SYNCHRO_TIMEOUT_MAX}] (seconds)')

    @staticmethod
    def to_ticks(value: str) -> int:
        """ Convert a string into a number of ticks, in [2;720].

        :param value: the number of ticks as a string
        :return: the number of ticks as an integer
        """
        try:
            ticks = integer(value)
            if SupvisorsOptions.INACTIVITY_TICKS_MIN > ticks or ticks > SupvisorsOptions.INACTIVITY_TICKS_MAX:
                raise ValueError
            return ticks
        except ValueError:
            raise ValueError(f'invalid value for inactivity_ticks: "{value}".'
                             f' integer expected in [{SupvisorsOptions.INACTIVITY_TICKS_MIN};'
                             f'{SupvisorsOptions.INACTIVITY_TICKS_MAX}]')

    @staticmethod
    def to_event_link(value: str) -> EventLinks:
        """ Convert a string into a EventLinks enum. """
        try:
            event_link = EventLinks[value.upper()]
        except KeyError:
            raise ValueError(f'invalid value for event_link: "{value}".'
                             f' expected in {[x.name for x in EventLinks]}')
        return event_link

    @staticmethod
    def to_conciliation_strategy(value: str) -> ConciliationStrategies:
        """ Convert a string into a ConciliationStrategies enum. """
        try:
            strategy = ConciliationStrategies[value.upper()]
        except KeyError:
            raise ValueError(f'invalid value for conciliation_strategy: "{value}".'
                             f' expected in {[x.name for x in ConciliationStrategies]}')
        return strategy

    @staticmethod
    def to_starting_strategy(value: str) -> StartingStrategies:
        """ Convert a string into a StartingStrategies enum. """
        try:
            strategy = StartingStrategies[value.upper()]
        except KeyError:
            raise ValueError(f'invalid value for starting_strategy: "{value}".'
                             f' expected in {[x.name for x in StartingStrategies]}')
        return strategy

    @staticmethod
    def to_statistics_type(value: str) -> Tuple[bool, bool]:
        """ Convert a string into a pair of boolean values to allow host and/or process statistics. """
        str_stats_types = list_of_strings(value)
        if len(str_stats_types) == 0:
            raise ValueError('invalid value for stats_enabled: <empty>')
        stats_types = []
        for val in str_stats_types:
            # first try to use the enumeration values
            try:
                stats_types.append(StatisticsTypes[val.upper()])
            except KeyError:
                # try the boolean version
                try:
                    stats_types.append(StatisticsTypes.ALL if boolean(val) else StatisticsTypes.OFF)
                except ValueError:
                    raise ValueError(f'invalid value for stats_enabled: "{value}".'
                                     f' expected in {[x.name for x in StatisticsTypes]}')
        return (StatisticsTypes.ALL in stats_types or StatisticsTypes.HOST in stats_types,
                StatisticsTypes.ALL in stats_types or StatisticsTypes.PROCESS in stats_types)

    @staticmethod
    def to_period(value: str) -> float:
        """ Convert a string into a list of period values. """
        try:
            period = float(value)
            if 1.0 > period or period > 3600.0:
                raise ValueError
            return period
        except ValueError:
            raise ValueError(f'invalid value for stats_collecting_period: "{value}".'
                             f' float expected in [1.0;3600.0] (seconds)')

    @staticmethod
    def to_periods(value: str) -> List[float]:
        """ Convert a string into a list of period values. """
        str_periods = list_of_strings(value)
        if len(str_periods) == 0:
            raise ValueError(f'unexpected number of stats_periods: {len(str_periods)}.'
                             ' minimum is 1')
        if len(str_periods) > 3:
            raise ValueError(f'unexpected number of stats_periods: {len(str_periods)}.'
                             ' maximum is 3')
        periods = []
        for val in str_periods:
            try:
                period = float(val)
                if 1.0 > period or period > 3600.0:
                    raise ValueError
                periods.append(period)
            except ValueError:
                raise ValueError(f'invalid value for stats_periods: "{val}".'
                                 f' float expected in [1.0;3600.0] (seconds)')
        return sorted(periods)

    @staticmethod
    def to_histo(value: str) -> int:
        """ Convert a string into a value of historic depth, in [10;1500].

        :param value: the historic size as a string
        :return: the historic size as an integer
        """
        try:
            histo = integer(value)
            if 10 > histo or histo > 1500:
                raise ValueError
            return histo
        except ValueError:
            raise ValueError(f'invalid value for stats_histo: "{value}".'
                             f' integer expected in [10;1500] (seconds)')


class SupvisorsServerOptions(ServerOptions):
    """ Class used to parse the options of the 'supvisors' section in the supervisor configuration file.

    Attributes are:
        - parser: the config parser ;
        - program_configs: the program configuration not retained by Supervisor ;
        - process_configs: the process configuration not retained by Supervisor.
    """

    def __init__(self, supvisors):
        """ Initialization of the attributes. """
        ServerOptions.__init__(self)
        self.supvisors = supvisors
        # attributes
        self.parser = None
        self.program_configs: Dict[str, ProgramConfig] = {}  # {program_name: ProgramConfig}
        self.process_configs: Dict[str, SupvisorsProcessConfig] = {}  # {process_name: SupvisorsProcessConfig}
        # disabilities for local processes (Supervisor issue #591)
        self.disabilities: Dict[str, bool] = {}  # {program_name: disability}
        self.read_disabilities()

    @property
    def logger(self) -> Logger:
        """ Get the Supvisors logger. """
        return self.supvisors.logger

    @property
    def supvisors_options(self) -> SupvisorsOptions:
        """ Get the Supvisors options. """
        return self.supvisors.options

    # Supervisor issue #591
    def read_disabilities(self) -> None:
        """ Read disabilities file and apply data to Supervisor processes.

        :return: None
        """
        self.disabilities = {}
        disabilities_file = self.supvisors_options.disabilities_file
        if disabilities_file:
            self.logger.info(f'SupvisorsServerOptions.read_disabilities: file={disabilities_file}')
            if os.path.isfile(disabilities_file):
                with open(disabilities_file) as in_file:
                    data = json.load(in_file)
                    if type(data) is dict:
                        self.disabilities = data
                        self.logger.debug(f'SupvisorsServerOptions.read_disabilities: disabilities={self.disabilities}')
            else:
                self.logger.debug(f'SupvisorsServerOptions.read_disabilities: {disabilities_file} not found')
        else:
            self.logger.warn('SupvisorsServerOptions.read_disabilities: no persistence for program disabilities')

    def write_disabilities(self) -> None:
        """ Write disabilities file from Supervisor processes.

        :return: None.
        """
        disabilities_file = self.supvisors_options.disabilities_file
        if disabilities_file:
            # serialize to the file defined in options
            with open(disabilities_file, 'w+') as out_file:
                out_file.write(json.dumps(self.disabilities))

    def enable_program(self, program_name: str) -> None:
        """ Re-enable the processes to be started and trigger their autostart if configured to.

        :param program_name: the program to enable
        :return: None
        """
        self.disabilities[program_name] = False
        self.write_disabilities()

    def disable_program(self, program_name: str) -> None:
        """ Disable the program so that the corresponding processes cannot be started.

        :param program_name: the program to disable
        :return: None
        """
        self.disabilities[program_name] = True
        self.write_disabilities()

    # Get additional information not stored by Supervisor when parsing the configuration files
    def _processes_from_section(self, parser, section: str, group_name: str, klass=None) -> List[ProcessConfig]:
        """ This method is overridden to store the configuration information not kept by Supervisor.

        This is originally used in Supervisor to set the real program name from the format defined in the ini file.
        However, Supervisor does not keep this information in its internal structure.

        :param parser: the config parser
        :param section: the program section
        :param group_name: the group that embeds the program definition
        :param klass: the ProcessConfig class (or EventListenerConfig or FastCGIProcessConfig)
        :return: the list of ProcessConfig
        """
        # keep a reference to the parser, so that it is not garbage-collected
        # it will be needed to re-evaluate procnums
        self.parser = parser
        # call super behaviour
        process_configs = ServerOptions._processes_from_section(self, parser, section, group_name, klass)
        # store process configurations and groups
        program_name = section.split(':', 1)[1]
        if program_name in self.program_configs:
            # get the existing program configuration
            program_config = self.program_configs[program_name]
        else:
            # create the program configuration
            program_config = ProgramConfig(program_name, klass)
            program_config.numprocs = int(parser.saneget(section, 'numprocs', '1'))
            program_config.disabled = self.disabilities.setdefault(program_name, False)
            self.program_configs[program_name] = program_config
        # store a generic process_config extension (identical for all groups)
        for idx, process_config in enumerate(process_configs):
            # process_config.name is the process_name (without group name)
            alt_process_config = self.process_configs.get(process_config.name)
            if not alt_process_config:
                # create a process configuration to get the associated program configuration
                alt_process_config = SupvisorsProcessConfig(program_config, idx, process_config.command)
                self.process_configs[process_config.name] = alt_process_config
        # associate the group to the program
        program_config.group_config_info[group_name] = process_configs
        # return super result
        return process_configs

    def get_section(self, program_name: str):
        """ Get the Supervisor relevant section name depending on the program name.

        :param program_name: the name of the program configured
        :return: the Supervisor section name
        """
        klass = self.program_configs[program_name].klass
        if klass is FastCGIProcessConfig:
            return f'fcgi-program:{program_name}'
        if klass is EventListenerConfig:
            return f'eventlistener:{program_name}'
        return f'program:{program_name}'

    def get_subprocesses(self, program_name) -> NameList:
        """ Return all processes related to the program definition.

        :param program_name: the name of the program, as declared in the configuration files.
        :return: all namespecs corresponding to the program.
        """
        program_groups: GroupConfigInfo = self.program_configs[program_name].group_config_info
        return [make_namespec(group_name, process_config.name)
                for group_name, process_config_list in program_groups.items()
                for process_config in process_config_list]

    def update_numprocs(self, program_name: str, numprocs: int) -> GroupConfigInfo:
        """ This method updates the program configuration based on the new numprocs value.

        :param program_name: the program name, as found in the sections of the Supervisor configuration files.
        :param numprocs: the new numprocs value.
        :return: the new process configurations.
        """
        section = self.get_section(program_name)
        # update the parser value
        self.logger.debug(f'SupvisorsServerOptions.update_numprocs: update parser section={section}'
                          f' with numprocs={numprocs}')
        self.parser[section]['numprocs'] = str(numprocs)
        # get the existing program configuration
        program = self.program_configs[program_name]
        program.numprocs = numprocs
        # rebuild the process configs from the new Supervisor configuration
        group_configs = {}
        for group_name, process_list in program.group_config_info.items():
            # remove the former process configuration
            for process in process_list:
                self.process_configs.pop(process.name, None)
            # build the new configuration
            group_configs[group_name] = self.processes_from_section(self.parser, section, group_name, program.klass)
        return group_configs



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import os

from supervisor.states import ProcessStates, RUNNING_STATES

from supvisors.instancestatus import SupvisorsInstanceStatus
from supvisors.statscompiler import ProcStatisticsInstance
from supvisors.ttypes import SupvisorsFaults, Payload, PayloadList
from supvisors.utils import get_small_value
from .viewcontext import *
from .viewinstance import SupvisorsInstanceView
from .webutils import *


class ProcInstanceView(SupvisorsInstanceView):
    """ View renderer of the Process section of the Supvisors Instance page.
    Inheritance is made from supervisor.web.StatusView to benefit from the action methods.
    Note that StatusView inheritance has been patched dynamically in supvisors.plugin.make_supvisors_rpcinterface
    so that StatusView inherits from ViewHandler instead of MeldView.
    """

    ProcessStats = Tuple[int, Optional[ProcStatisticsInstance]]

    def __init__(self, context):
        """ Call of the superclass constructors. """
        SupvisorsInstanceView.__init__(self, context, PROC_INSTANCE_PAGE)

    def handle_parameters(self):
        """ Retrieve the parameters selected on the web page. """
        super().handle_parameters()
        # pre-fill the message here to warn the user about the actions on this page
        # it will always be displayed by default unless it is overwritten by another message
        self.view_ctx.set_message_default('The Supvisors rules do NOT apply here', Warn)

    def write_options(self, header_elt):
        """ Write configured periods for statistics.
        Does nothing by default. To be specialized in subclasses where statistics are available. """
        if not (self.has_process_statistics or self.has_host_statistics):
            # hide the whole card and the following line
            header_elt.findmeld('option_card_mid').replace('')
            header_elt.findmeld('option_line_mid').replace('')
        else:
            # configure the process periods
            if self.has_process_statistics:
                self.write_periods(header_elt)
            else:
                # hide the Statistics periods box
                header_elt.findmeld('period_div_mid').replace('')
            # configure the view switch
            if self.has_host_statistics:
                self.write_view_switch(header_elt)
            else:
                # remove whole box if statistics are disabled
                # Host page is useless in this case
                header_elt.findmeld('view_div_mid').replace('')

    def write_view_switch(self, header_elt):
        """ Configure the statistics view buttons. """
        elt = header_elt.findmeld('host_view_a_mid')
        elt.content(f'{self.sup_ctx.local_status.supvisors_id.host_id}')
        url = self.view_ctx.format_url('', HOST_INSTANCE_PAGE)
        elt.attributes(href=url)

    # RIGHT SIDE / BODY part
    def write_contents(self, contents_elt):
        """ Rendering of the contents part of the page. """
        sorted_data, excluded_data = self.get_process_data()
        self.write_process_table(contents_elt, sorted_data, excluded_data)
        # check selected Process Statistics
        namespec = self.view_ctx.process_name
        if namespec:
            if not self.has_process_statistics:
                # statistics are not available for this Supvisors instance
                self.view_ctx.process_name = ''
            elif namespec != 'supervisord':
                # unselect if not running in this Supvisors instance
                status = self.view_ctx.get_process_status(namespec)
                if not status or self.view_ctx.local_identifier not in status.running_identifiers:
                    self.logger.warn(f'ProcInstanceView.write_contents: unselect Process Statistics for {namespec}')
                    # former parameter is not consistent with local instance. remove it
                    self.view_ctx.process_name = ''
        # write selected Process Statistics
        namespec = self.view_ctx.process_name
        info = next(filter(lambda x: x['namespec'] == namespec, sorted_data + excluded_data), {})
        self.write_process_statistics(contents_elt, info)

    def get_process_data(self) -> Tuple[PayloadList, PayloadList]:
        """ Collect sorted data on processes.
        The sorted data are meant to be displayed in the process table.
        The excluded data are not but may be used for process statistics selection.

        :return: the sorted data and the excluded data.
        """
        # extract what is useful to display
        local_identifier = self.view_ctx.local_identifier
        status: SupvisorsInstanceStatus = self.sup_ctx.instances[local_identifier]
        data = []
        for namespec, process in status.processes.items():
            # a 'main' process has the same name as its namespec
            main = process.process_name == namespec
            info = process.info_map[self.view_ctx.local_identifier]
            crashed = ProcessStatus.is_crashed_event(info)
            proc_stats = self.view_ctx.get_process_stats(namespec, local_identifier)
            payload = {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
                       'application_name': info['group'], 'process_name': info['name'], 'namespec': namespec,
                       'main': main, 'identifier': local_identifier,
                       'disabled': info['disabled'], 'startable': not info['disabled'], 'stoppable': True,
                       'statename': info['statename'], 'statecode': info['state'],
                       'gravity': 'FATAL' if crashed else info['statename'],
                       'has_crashed': info['has_crashed'],
                       'description': info['description'],
                       'expected_load': process.rules.expected_load,
                       'proc_stats': proc_stats,
                       'has_stdout': process.has_stdout(local_identifier),
                       'has_stderr': process.has_stderr(local_identifier)}
            data.append(payload)
        # re-arrange data
        sorted_data, excluded_data = self.sort_data(data)
        # add supervisord payload at the end of sorted data
        sorted_data.append(self.get_supervisord_data(status))
        return sorted_data, excluded_data

    def get_supervisord_data(self, status: SupvisorsInstanceStatus) -> Payload:
        """ Collect sorted data on supervisord process.

        :param status: the local Supvisors instance
        :return: the supervisord data.
        """
        local_identifier = self.view_ctx.local_identifier
        proc_stats = self.view_ctx.get_process_stats('supervisord', local_identifier)
        payload = {'row_type': ProcessRowTypes.SUPERVISOR_PROCESS,
                   'application_name': 'supervisord', 'process_name': 'supervisord', 'namespec': 'supervisord',
                   'main': True, 'identifier': local_identifier,
                   'disabled': False, 'startable': False, 'stoppable': True,
                   'statename': 'RUNNING', 'statecode': 20,
                   'gravity': 'RUNNING', 'has_crashed': False,
                   'expected_load': 0, 'proc_stats': proc_stats}
        # add description (pid / uptime) as done by Supervisor
        info = {'state': ProcessStates.RUNNING, 'start': status.times.start_local_mtime,
                'now': status.times.local_mtime, 'pid': os.getpid()}
        payload['description'] = ProcessStatus.update_description(info)
        return payload

    def sort_data(self, data: PayloadList) -> Tuple[PayloadList, PayloadList]:
        """ This method sorts a process list by application and using the alphabetical order.
        Processes belonging to an application may be removed, depending on the shex user selection.

        :param data: a list of process payloads
        :return: the sorted list and the excluded list.
        """
        sorted_data, excluded_data = [], []
        # sort processes by application
        application_map = {}
        for info in data:
            application_map.setdefault(info['application_name'], []).append(info)
        # sort applications alphabetically
        for application_name, application_processes in sorted(application_map.items()):
            main = len(application_processes) == 1 and application_processes[0]['main']
            application_shex = True
            if not main:
                # add entry for application
                sorted_data.append(self.get_application_summary(application_name, application_processes))
                # filter data depending on their application shex
                application_shex, _ = self.view_ctx.get_application_shex(application_name)
            if application_shex:
                # add processes using the alphabetical ordering
                sorted_list = sorted([info for info in application_processes],
                                     key=lambda x: x['process_name'])
                sorted_data.extend(sorted_list)
            else:
                # push to excluded data
                excluded_data.extend(application_processes)
        return sorted_data, excluded_data

    def get_application_summary(self, application_name: str, application_processes: PayloadList) -> Payload:
        """ Get a summary of the application based on a subset of its processes.

        :param application_name: the name of the application
        :param application_processes: the subset of the application processes running on the same node
        :return: the application payload to be displayed
        """
        expected_load, appli_stats = self.sum_process_info(application_processes)
        # create application payload
        application = self.sup_ctx.applications[application_name]
        payload = {'row_type': ProcessRowTypes.APPLICATION,
                   'application_name': application_name, 'process_name': None, 'namespec': None,
                   'identifier': self.view_ctx.local_identifier,
                   'disabled': False, 'startable': False, 'stoppable': True,
                   'statename': application.state.name, 'statecode': application.state.value,
                   'gravity': application.state.name, 'has_crashed': False,
                   'description': application.get_operational_status(),
                   'nb_items': len(application_processes), 'expected_load': expected_load,
                   'proc_stats': appli_stats}
        return payload

    @staticmethod
    def sum_process_info(data: PayloadList) -> ProcessStats:
        """ Get the total resources taken by the processes.

        :param data: the list of process payloads
        :return: the total expected load, number of processor cores, memory and CPU
        """
        expected_load, cpu, mem = 0, 0, 0
        reset = True
        for info in data:
            if info['statecode'] in RUNNING_STATES:
                expected_load += info['expected_load']
                # sum CPU / Mem stats
                proc_stats = info['proc_stats']
                if proc_stats:
                    if len(proc_stats.cpu) and len(proc_stats.mem):
                        reset = False
                        # the most recent value is at the end of the list
                        cpu += proc_stats.cpu[-1]
                        mem += proc_stats.mem[-1]
        # reset appli_stats if no process involved
        # keep output similar to process stats
        appli_stats = None
        if not reset:
            appli_stats = ProcStatisticsInstance()
            appli_stats.cpu = [cpu]
            appli_stats.mem = [mem]
        return expected_load, appli_stats

    def write_process_table(self, contents_elt, sorted_data: PayloadList, excluded_data: PayloadList) -> None:
        """ Rendering of the processes managed in Supervisor.

        :param contents_elt: the root element of the page.
        :param sorted_data: the process data displayed.
        :param excluded_data: the process data not displayed.
        :return: None.
        """
        table_elt = contents_elt.findmeld('table_mid')
        if sorted_data:
            self.write_application_global_shex(table_elt)
            # remove stats columns if statistics are disabled
            self.write_common_process_table(table_elt)
            # loop on all processes
            shaded_appli_tr, shaded_proc_tr = False, False  # used to invert background style
            for tr_elt, info in table_elt.findmeld('tr_mid').repeat(sorted_data):
                if info['row_type'] == ProcessRowTypes.INSTANCE_PROCESS:
                    # this is a process row
                    if info['main']:
                        # write common status (shared between this process view and application view)
                        self.write_common_process_status(tr_elt, info)
                        # main row background follows the same logic as applications
                        apply_shade(tr_elt, shaded_appli_tr)
                        shaded_appli_tr = not shaded_appli_tr
                    else:
                        # remove shex td
                        tr_elt.findmeld('shex_td_mid').replace('')
                        # write common status (shared between this process view and application view)
                        self.write_common_process_status(tr_elt, info)
                        # set line background and invert
                        apply_shade(tr_elt, shaded_proc_tr)
                        shaded_proc_tr = not shaded_proc_tr
                elif info['row_type'] == ProcessRowTypes.SUPERVISOR_PROCESS:
                    # supervisord row has a slightly different format
                    self.write_supervisord_status(tr_elt, info)
                    # set line background and invert
                    apply_shade(tr_elt, shaded_appli_tr)
                    shaded_appli_tr = not shaded_appli_tr
                elif info['row_type'] == ProcessRowTypes.APPLICATION:
                    # this is an application row
                    self.write_application_status(tr_elt, info, shaded_appli_tr)
                    # set line background and invert
                    apply_shade(tr_elt, shaded_appli_tr)
                    shaded_appli_tr = not shaded_appli_tr
                    shaded_proc_tr = shaded_appli_tr
            # writes the total statistics line
            self.write_total_status(table_elt, sorted_data, excluded_data)
        else:
            table_elt.replace('No programs to display')

    def write_application_global_shex(self, table_elt) -> None:
        """ Write global shrink / expand buttons.

        :param table_elt: the process table element.
        :return: None.
        """
        shex = self.view_ctx.application_shex
        expand_shex = self.view_ctx.get_default_application_shex(True)
        shrink_shex = self.view_ctx.get_default_application_shex(False)
        self.write_global_shex(table_elt, APP_SHRINK_EXPAND, shex, expand_shex, shrink_shex)

    def write_application_status(self, tr_elt, info, shaded_tr):
        """ Write the application section into a table. """
        application_name = info['application_name']
        # print application shex
        elt = tr_elt.findmeld('shex_td_mid')
        application_shex, inverted_shex = self.view_ctx.get_application_shex(application_name)
        if application_shex:
            elt.attrib['rowspan'] = str(info['nb_items'] + 1)
            apply_shade(elt, shaded_tr)
        elt = elt.findmeld('shex_a_mid')
        self.logger.trace(f'ProcInstanceView.write_application_status: application_name={application_name}'
                          f' application_shex={application_shex} inverted_shex={inverted_shex}')
        elt.content(f'{SHEX_SHRINK if application_shex else SHEX_EXPAND}')
        url = self.view_ctx.format_url('', self.page_name, **{APP_SHRINK_EXPAND: inverted_shex})
        elt.attributes(href=url)
        # print application name (covers state and description)
        elt = tr_elt.findmeld('name_td_mid')
        elt.attrib['colspan'] = '3'
        elt = elt.findmeld('name_a_mid')
        elt.content(application_name)
        url = self.view_ctx.format_url('', APPLICATION_PAGE, **{APPLI: application_name})
        elt.attributes(href=url)
        for mid in ['state_td_mid', 'desc_td_mid']:
            tr_elt.findmeld(mid).replace('')
        # print application statistics
        self.write_common_statistics(tr_elt, info)
        # start / stop / restart group actions
        self._write_process_button(tr_elt, 'start_a_mid', '', self.page_name, 'startgroup', f'{application_name}:')
        self._write_process_button(tr_elt, 'stop_a_mid', '', self.page_name, 'stopgroup', f'{application_name}:')
        self._write_process_button(tr_elt, 'restart_a_mid', '', self.page_name, 'restartgroup', f'{application_name}:')
        # remove log actions
        elt = tr_elt.findmeld('clear_td_mid')
        elt.attrib['colspan'] = '3'
        elt.content('')
        for mid in ['tailout_td_mid', 'tailerr_td_mid']:
            tr_elt.findmeld(mid).replace('')

    def write_supervisord_status(self, tr_elt, info: Payload) -> None:
        """ Write the supervisord status into a table. """
        # display Master symbol in shex column
        if self.sup_ctx.is_master:
            elt = tr_elt.findmeld('shex_td_mid')
            elt.content(MASTER_SYMBOL)
        # print process name
        elt = tr_elt.findmeld('name_a_mid')
        elt.content(info['process_name'])
        url = self.view_ctx.format_url('', MAIN_TAIL_PAGE, **{PROCESS: info['namespec'],
                                                              LIMIT: self.supvisors.options.tail_limit})
        elt.attributes(href=url, target="_blank")
        # print common status
        self.write_common_state(tr_elt, info)
        self.write_common_statistics(tr_elt, info)
        # manage actions (start not applicable)
        self._write_supervisord_off_button(tr_elt, 'start_a_mid')
        self._write_supervisord_button(tr_elt, 'stop_a_mid', self.page_name, **{ACTION: 'shutdownsup'})
        self._write_supervisord_button(tr_elt, 'restart_a_mid', self.page_name, **{ACTION: 'restartsup'})
        # manage log actions (stderr not applicable)
        self._write_supervisord_button(tr_elt, 'clear_a_mid', self.page_name, **{ACTION: 'mainclearlog'})
        self._write_supervisord_button(tr_elt, 'tailout_a_mid', MAIN_STDOUT_PAGE)
        self._write_supervisord_off_button(tr_elt, 'tailerr_a_mid')

    def _write_supervisord_button(self, tr_elt, mid: str, page: str, **action) -> None:
        """ Write the configuration of the button of supervisord. """
        elt = tr_elt.findmeld(mid)
        update_attrib(elt, 'class', 'button on')
        url = self.view_ctx.format_url('', page, **action)
        elt.attributes(href=url)

    @staticmethod
    def _write_supervisord_off_button(tr_elt, mid: str) -> None:
        """ Write the configuration of the button of supervisord. """
        elt = tr_elt.findmeld(mid)
        update_attrib(elt, 'class', 'button off')

    def write_total_status(self, table_elt, sorted_data: PayloadList, excluded_data: PayloadList):
        """ Write the total statistics for this Supvisors instance.

        :param table_elt: the table element
        :param sorted_data: the process data displayed
        :param excluded_data: the process data not displayed
        :return: None
        """
        tr_elt = table_elt.findmeld('total_mid')
        # sum MEM and CPU stats of all processes
        expected_load, appli_stats = self.sum_process_info(sorted_data + excluded_data)
        # update Load
        tr_elt.findmeld('load_total_th_mid').content(f'{expected_load}')
        if appli_stats:
            # update MEM & CPU
            tr_elt.findmeld('mem_total_th_mid').content(get_small_value(appli_stats.mem[0]))
            tr_elt.findmeld('cpu_total_th_mid').content(get_small_value(appli_stats.cpu[0]))

    # ACTION part
    def make_callback(self, namespec, action):
        """ Triggers processing iaw action requested """
        if action == 'startgroup':
            return self.start_group_action(namespec)
        if action == 'stopgroup':
            return self.stop_group_action(namespec)
        if action == 'restartgroup':
            return self.restart_group_action(namespec)
        if action == 'mainclearlog':
            return self.clear_log_action()
        result = super().make_callback(namespec, action)
        # WARN: this ugly part is necessary to handle the DISABLED exception that can be raised from Supervisor
        #       startProcess. It is not possible to patch Supervisor make_callback without copying a huge piece
        #       of source code from Supervisor, so here is an inspection of the make_callback result to check
        #       for a callable declaring an 'unexpected rpc fault [103]'
        if callable(result):
            message = result()
            if type(message) is str and f'[{SupvisorsFaults.DISABLED.value}]' in message:
                return delayed_error(f'Process {namespec}: disabled')
        return result

    def start_group_action(self, namespec: str) -> Callable:
        """ Start all processes in the group.
        The RPC wait parameter is linked to the auto-refresh parameter of the page.

        :param namespec: the group processes that have to be started (expecting a form like 'group:*').
        :return: a callable for deferred result.
        """
        self.logger.debug(f'ProcInstanceView.start_group_action: group_name={namespec}')
        wait = not self.view_ctx.auto_refresh
        return self.supervisor_rpc_action('startProcess', (f'{namespec}', wait), f'Group {namespec} started')

    def stop_group_action(self, namespec: str) -> Callable:
        """ Stop all processes in the group.
        The RPC wait parameter is linked to the auto-refresh parameter of the page.

        :param namespec: the group processes that have to be stopped (expecting a form like 'group:*').
        :return: a callable for deferred result.
        """
        self.logger.debug(f'ProcInstanceView.stop_group_action: group_name={namespec}')
        wait = not self.view_ctx.auto_refresh
        return self.supervisor_rpc_action('stopProcess', (f'{namespec}', wait), f'Group {namespec} stopped')

    def restart_group_action(self, namespec: str) -> Callable:
        """ Start all processes in the group.
        The RPC wait parameter is linked to the auto-refresh parameter of the page and set only on the last call.

        :param namespec: the group processes that have to be restarted (expecting a form like 'group:*').
        :return: a callable for deferred result.
        """
        self.logger.debug(f'ProcInstanceView.restart_group_action: group_name={namespec}')
        wait = not self.view_ctx.auto_refresh
        multicall = [{'methodName': 'supervisor.stopProcess', 'params': [f'{namespec}']},
                     {'methodName': 'supervisor.startProcess', 'params': [f'{namespec}', wait]}]
        return self.multicall_rpc_action(multicall, f'Group {namespec} restarted')

    def clear_log_action(self) -> Callable:
        """ Clear the log of Supervisor.

        :return: a callable for deferred result
        """
        return self.supervisor_rpc_action('clearLog', (), 'Log for Supervisor cleared')



# ======================================================================
# Copyright 2017 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from unittest.mock import Mock

import pytest

from supvisors.initializer import *
from supvisors.statscollector import StatisticsCollectorProcess
from .base import DummySupervisor


def test_creation(mocker):
    """ Test the values set at construction. """
    mocked_parser = mocker.patch('supvisors.initializer.Parser', return_value='Parser')
    mocked_srv_options = Mock(procnumbers={})
    mocked_options = mocker.patch('supvisors.initializer.SupvisorsServerOptions', return_value=mocked_srv_options)
    # create the instance to test, using default empty configuration
    supv = Supvisors(DummySupervisor())
    # test calls
    assert mocked_options.called
    assert mocked_parser.called
    # test instances
    assert supv.rpc_handler is None
    assert supv.discovery_handler is None
    assert supv.external_publisher is None
    assert isinstance(supv.options, SupvisorsOptions)
    assert mocked_srv_options.realize.called
    assert isinstance(supv.logger, Logger)
    assert isinstance(supv.supervisor_data, SupervisorData)
    assert isinstance(supv.supervisor_updater, SupervisorUpdater)
    assert isinstance(supv.mapper, SupvisorsMapper)
    assert isinstance(supv.stats_collector, StatisticsCollectorProcess)
    assert isinstance(supv.context, Context)
    assert isinstance(supv.starter, Starter)
    assert isinstance(supv.starter_model, StarterModel)
    assert isinstance(supv.stopper, Stopper)
    assert isinstance(supv.host_compiler, HostStatisticsCompiler)
    assert isinstance(supv.process_compiler, ProcStatisticsCompiler)
    assert isinstance(supv.fsm, FiniteStateMachine)
    assert supv.parser == 'Parser'
    assert isinstance(supv.listener, SupervisorListener)


def test_create_logger():
    """ Test the create_logger method. """
    # create Supvisors instance
    supervisor = DummySupervisor()
    # test AUTO logfile
    logger_config = get_logger_configuration()
    logger_config['logfile'] = Automatic
    assert create_logger(supervisor, logger_config) is supervisor.options.logger
    assert supervisor.options.logger.handlers[0].fmt == LOGGER_FORMAT
    # test defined logfile
    logger_config['prefix'] = 'Supvisors'
    logger_config['logfile'] = '/tmp/dummy.log'
    logger = create_logger(supervisor, logger_config)
    assert logger is not supervisor.options.logger
    assert supervisor.options.logger.handlers[0].fmt == LOGGER_FORMAT
    assert logger.handlers[0].fmt == f'Supvisors;{LOGGER_FORMAT}'


def test_identifier_exception(mocker):
    """ Test the values set at construction. """
    mocker.patch('supvisors.initializer.SupvisorsServerOptions')
    mocker.patch('supvisors.initializer.SupvisorsMapper.configure', side_effect=ValueError)
    # create Supvisors instance
    supervisord_instance = DummySupervisor()
    # test that local node exception raises a failure to Supervisor
    with pytest.raises(ValueError):
        Supvisors(supervisord_instance)


def test_psutil_exception(mocker):
    """ Test the values set at construction. """
    mocker.patch('supvisors.initializer.SupvisorsServerOptions')
    mocker.patch.dict('sys.modules', {'supvisors.statscollector': None})
    # create Supvisors instance
    supvisors = Supvisors(DummySupervisor())
    # test that parser exception is accepted
    assert supvisors.stats_collector is None


def test_parser_exception(mocker):
    """ Test the values set at construction. """
    mocker.patch('supvisors.initializer.Parser', side_effect=Exception)
    mocker.patch('supvisors.initializer.SupvisorsServerOptions')
    # create Supvisors instance
    supvisors = Supvisors(DummySupervisor())
    # test that parser exception is accepted
    assert supvisors.parser is None


def test_supvisors_init():
    """ Just import supvisors to test __init__.py file. """
    import supvisors
    assert supvisors.__name__ == 'supvisors'



# ======================================================================
# Copyright 2017 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from unittest.mock import call, Mock

import pytest
from supervisor.rpcinterface import SupervisorNamespaceRPCInterface
from supervisor.supervisorctl import DefaultControllerPlugin
from supervisor.xmlrpc import Faults

from supvisors import __version__
from supvisors.supvisorsctl import *
from supvisors.ttypes import SupvisorsFaults


@pytest.fixture
def remote_proxy(mocker):
    """ Patch the remote proxy getter. """

    def mock_server_proxy(uri, transport):
        return Mock(uri=uri, transport=transport,
                    supervisor=Mock(spec=SupervisorNamespaceRPCInterface),
                    supvisors=Mock(spec=RPCInterface))

    return mocker.patch('supvisors.supvisorsctl.xmlrpclib.ServerProxy', side_effect=mock_server_proxy)


@pytest.fixture
def controller(remote_proxy):
    """ Patch the controller instance. """
    controller = Mock(spec=Controller, exitstatus=LSBInitExitStatuses.SUCCESS)
    controller.get_server_proxy.return_value = Mock(spec=RPCInterface)
    controller.options = Mock(username='cliche',
                              password='p@$$w0rd',
                              serverurl='dummy_url',
                              plugins=[DefaultControllerPlugin(controller)])
    return controller


@pytest.fixture
def plugin(controller):
    """ Create the instance to test. """
    return ControllerPlugin(controller)


@pytest.fixture
def mocked_check(mocker):
    """ Mock _upcheck. """
    return mocker.patch('supvisors.supvisorsctl.ControllerPlugin._upcheck', return_value=True)


def _check_output_error(controller, error):
    """ Test output error of controller. """
    assert controller.output.called
    assert any('ERROR' in str(ocall) for ocall in controller.output.call_args_list) == error
    assert not error or controller.exitstatus != LSBInitExitStatuses.SUCCESS
    controller.output.reset_mock()


def _check_call(controller, mocked_check, mocked_rpc, help_fct, do_fct, arg, rpc_result, has_error=False):
    """ Generic test of help and request. """
    # 1. test that help uses output
    help_fct()
    assert controller.output.called
    controller.output.reset_mock()
    # 2. test normal request
    do_fct(arg)
    # check upcheck call if any
    if mocked_check:
        assert mocked_check.call_args_list == [call()]
        mocked_check.reset_mock()
    # check RPC
    assert mocked_rpc.call_args_list == rpc_result
    mocked_rpc.reset_mock()
    # check output (with no error)
    _check_output_error(controller, has_error)
    # 3. test request error
    mocked_rpc.side_effect = xmlrpclib.Fault(0, 'error')
    do_fct(arg)
    mocked_rpc.side_effect = None
    # check upcheck call if any
    if mocked_check:
        assert mocked_check.call_args_list == [call()]
        mocked_check.reset_mock()
    # check RPC
    assert mocked_rpc.call_args_list == rpc_result
    mocked_rpc.reset_mock()
    # check output (with error)
    _check_output_error(controller, True)


def _check_start_application_command(controller, mocked_check, mocked_appli, mocked_rpc,
                                     help_cmd, do_cmd, all_results, sel_args, sel_results):
    """ Common test of a starting command. """
    # test the request using few arguments
    do_cmd('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using unknown strategy
    do_cmd('strategy')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test request to start all
    mocked_appli.return_value = [{'application_name': 'appli_1', 'managed': True},
                                 {'application_name': 'appli_2', 'managed': True},
                                 {'application_name': 'appli_3', 'managed': False}]
    # first possibility: use no name
    rpc_result = [call(1, result) for result in all_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, 'LESS_LOADED', rpc_result)
    assert mocked_appli.call_args_list == [call(), call()]
    mocked_appli.reset_mock()
    # second possibility: use 'all'
    rpc_result = [call(2, result) for result in all_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, 'MOST_LOADED all', rpc_result)
    assert mocked_appli.call_args_list == [call(), call()]
    mocked_appli.reset_mock()
    # test help and request for starting a selection
    rpc_result = [call(0, result) for result in sel_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, 'CONFIG ' + sel_args, rpc_result, True)
    # test help and request with get_all_applications_info error
    mocked_appli.reset_mock()
    mocked_appli.side_effect = xmlrpclib.Fault(0, 'error')
    do_cmd('LESS_LOADED')
    assert mocked_appli.call_args_list == [call()]
    assert not mocked_rpc.called
    _check_output_error(controller, True)


def _check_start_process_command(controller, mocked_check, mocked_info, mocked_rpc,
                                 help_cmd, do_cmd, all_results, sel_args, sel_results):
    """ Common test of a starting command. """
    # test the request using few arguments
    do_cmd('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using unknown strategy
    do_cmd('strategy')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test request to start all
    mocked_info.return_value = [{'application_name': 'appli_1', 'process_name': 'proc_1'},
                                {'application_name': 'appli_2', 'process_name': 'proc_3'}]
    # first possibility: use no name
    rpc_result = [call(1, result) for result in all_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, 'LESS_LOADED', rpc_result)
    assert mocked_info.call_args_list == [call(), call()]
    mocked_info.reset_mock()
    # second possibility: use 'all'
    rpc_result = [call(2, result) for result in all_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, 'MOST_LOADED all', rpc_result)
    assert mocked_info.call_args_list == [call(), call()]
    mocked_info.reset_mock()
    # test help and request for starting a selection
    rpc_result = [call(0, result) for result in sel_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, 'CONFIG ' + sel_args, rpc_result)
    # test help and request with get_all_applications_info error
    mocked_info.reset_mock()
    mocked_info.side_effect = xmlrpclib.Fault(0, 'error')
    do_cmd('LESS_LOADED')
    assert mocked_info.call_args_list == [call()]
    assert not mocked_rpc.called
    _check_output_error(controller, True)


def _check_stop_application_command(controller, mocked_check, mocked_appli, mocked_rpc,
                                    help_cmd, do_cmd, all_results, sel_args, sel_results):
    """ Common test of a stopping command. """
    # test request to stop all
    mocked_appli.return_value = [{'application_name': 'appli_1', 'managed': True},
                                 {'application_name': 'appli_2', 'managed': True},
                                 {'application_name': 'appli_3', 'managed': False}]
    # first possibility: use no name
    rpc_result = [call(result) for result in all_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, '', rpc_result)
    assert mocked_appli.call_args_list == [call(), call()]
    mocked_appli.reset_mock()
    # second possibility: use 'all'
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, 'all', rpc_result)
    assert mocked_appli.call_args_list == [call(), call()]
    mocked_appli.reset_mock()
    # test help and request for starting a selection of applications
    rpc_result = [call(result) for result in sel_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, sel_args, rpc_result, True)
    # test help and request with get_all_applications_info error
    mocked_appli.reset_mock()
    mocked_appli.side_effect = xmlrpclib.Fault(0, 'error')
    do_cmd('')
    assert mocked_appli.call_args_list == [call()]
    assert mocked_rpc.call_count == 0
    _check_output_error(controller, True)


def _check_stop_process_command(controller, mocked_check, mocked_info, mocked_rpc,
                                help_cmd, do_cmd, all_results, sel_args, sel_results):
    """ Common test of a stopping command. """
    # test request to stop all
    mocked_info.return_value = [{'application_name': 'appli_1', 'process_name': 'proc_1'},
                                {'application_name': 'appli_2', 'process_name': 'proc_3'}]
    # first possibility: use no name
    rpc_result = [call(result) for result in all_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, '', rpc_result)
    assert mocked_info.call_args_list == [call(), call()]
    mocked_info.reset_mock()
    # second possibility: use 'all'
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, 'all', rpc_result)
    assert mocked_info.call_args_list == [call(), call()]
    mocked_info.reset_mock()
    # test help and request for starting a selection of applications
    rpc_result = [call(result) for result in sel_results]
    _check_call(controller, mocked_check, mocked_rpc, help_cmd, do_cmd, sel_args, rpc_result)
    # test help and request with get_all_applications_info error
    mocked_info.reset_mock()
    mocked_info.side_effect = xmlrpclib.Fault(0, 'error')
    do_cmd('')
    assert mocked_info.call_args_list == [call()]
    assert mocked_rpc.call_count == 0
    _check_output_error(controller, True)


def test_creation(controller, plugin):
    """ Test the creation of the Supvisors ControllerPlugin and the _startresult patch applied to the Supervisor
    DefaultControllerPlugin. """
    assert plugin.ctl is controller
    # check that the Supervisor Faults have been expanded
    assert Faults.DISABLED == SupvisorsFaults.DISABLED.value
    # check that the Supervisor plugin has been patched and test the patch
    supervisor_plugin = controller.options.plugins[0]
    result = {'group': 'dummy_group', 'name': 'dummy_process', 'status': Faults.DISABLED}
    assert supervisor_plugin._startresult(result) == 'dummy_group:dummy_process: ERROR disabled'
    result['status'] = Faults.SUCCESS
    assert supervisor_plugin._startresult(result) == 'dummy_group:dummy_process: started'


def test_get_server_proxy(remote_proxy, plugin):
    """ Test the access to any remote proxy. """
    # test the proxy
    assert SupervisorNamespaceRPCInterface == plugin.get_server_proxy('http://localhost:9000', 'supervisor')._spec_class
    assert RPCInterface == plugin.get_server_proxy('http://localhost:9000', 'supvisors')._spec_class
    assert remote_proxy.call_count == 2
    for called in remote_proxy.call_args_list:
        args, kwargs = called
        assert args[0] == 'http://127.0.0.1'
        assert isinstance(args[1], xmlrpc.SupervisorTransport)
        assert args[1].username == 'cliche'
        assert args[1].password == 'p@$$w0rd'
        assert args[1].serverurl == 'http://localhost:9000'


def test_supvisors(controller, plugin):
    """ Test the access to Supvisors proxy. """
    # test the proxy
    assert RPCInterface == plugin.supvisors()._spec_class
    assert controller.get_server_proxy.call_args_list == [call('supvisors')]


def test_get_running_instances(controller, plugin, mocked_check):
    """ Test the get_running_instances request. """
    mocked_rpc = plugin.supvisors().get_all_instances_info
    mocked_rpc.return_value = [{'identifier': 'third', 'node_name': '10.0.0.3', 'port': 30000, 'statecode': 3},
                               {'identifier': 'first', 'node_name': '10.0.0.1', 'port': 30000, 'statecode': 2},
                               {'identifier': 'second', 'node_name': '10.0.0.2', 'port': 60000, 'statecode': 0}]
    # test request
    assert plugin.get_running_instances() == {'first': 'http://10.0.0.1:30000', 'third': 'http://10.0.0.3:30000'}
    # test output (with no error)
    assert not controller.output.called
    # test request error
    mocked_rpc.side_effect = xmlrpclib.Fault(0, 'error')
    assert plugin.get_running_instances() == {}
    mocked_rpc.side_effect = None
    # test output (with error)
    _check_output_error(controller, True)


def test_sversion(controller, plugin, mocked_check):
    """ Test the sversion request. """
    mocked_rpc = plugin.supvisors().get_api_version
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_sversion, plugin.do_sversion, '', [call()])


def test_master(controller, plugin, mocked_check):
    """ Test the master request. """
    mocked_rpc = plugin.supvisors().get_master_identifier
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_master, plugin.do_master, '', [call()])


def test_strategies(controller, plugin, mocked_check):
    """ Test the master request. """
    mocked_rpc = plugin.supvisors().get_strategies
    mocked_rpc.return_value = {'conciliation': 'hard', 'starting': 'easy', 'auto-fencing': True}
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_strategies, plugin.do_strategies, '', [call()])


def test_statistics_status(controller, plugin, mocked_check):
    """ Test the master request. """
    mocked_rpc = plugin.supvisors().get_statistics_status
    mocked_rpc.return_value = {'host_stats': True, 'process_stats': False, 'collecting_period': 7.5}
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_stats_status, plugin.do_stats_status, '', [call()])


def test_sstate(controller, plugin, mocked_check):
    """ Test the sstate request. """
    mocked_rpc = plugin.supvisors().get_supvisors_state
    mocked_rpc.return_value = {'fsm_statecode': 10, 'fsm_statename': 'running',
                               'discovery_mode': True, 'master_identifier': '10.0.0.1',
                               'starting_jobs': [], 'stopping_jobs': ['10.0.0.1', 'test']}
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_sstate, plugin.do_sstate, '', [call()])


def test_instance_status(controller, plugin, mocked_check):
    """ Test the instance_status request. """
    mocked_rpc = plugin.supvisors().get_all_instances_info
    mocked_rpc.return_value = [{'nick_identifier': '10.0.0.1', 'identifier': '10.0.0.1:60000',
                                'node_name': '10.0.0.1', 'port': 60000,
                                'statename': 'running', 'discovery_mode': False,
                                'loading': 10, 'local_time': 1500, 'remote_sequence_counter': 12,
                                'process_failure': False,
                                'fsm_statename': 'OPERATION', 'master_identifier': '10.0.0.1',
                                'starting_jobs': True, 'stopping_jobs': False},
                               {'nick_identifier': '10.0.0.2', 'identifier': '10.0.0.2:60000',
                                'node_name': '10.0.0.2', 'port': 60000,
                                'statename': 'stopped', 'discovery_mode': True,
                                'loading': 0, 'local_time': 100, 'remote_sequence_counter': 15,
                                'process_failure': True,
                                'fsm_statename': 'CONCILATION', 'master_identifier': 'hostname',
                                'starting_jobs': False, 'stopping_jobs': True}]
    _check_call(controller, mocked_check, mocked_rpc,  plugin.help_instance_status, plugin.do_instance_status,
                '', [call()])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_instance_status, plugin.do_instance_status,
                'all', [call()])
    # test help and request for node status from a selection of address names
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_instance_status, plugin.do_instance_status,
                '10.0.0.2 10.0.0.1', [call()])


def test_application_info(controller, plugin, mocked_check):
    """ Test the application_info request. """
    mocked_rpc = plugin.supvisors().get_all_applications_info
    mocked_rpc.return_value = [{'application_name': 'appli_1', 'statename': 'running',
                                'major_failure': True, 'minor_failure': False},
                               {'application_name': 'appli_2', 'statename': 'stopped',
                                'major_failure': False, 'minor_failure': True}]
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_application_info, plugin.do_application_info,
                '', [call()])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_application_info, plugin.do_application_info,
                'all', [call()])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_application_info, plugin.do_application_info,
                'appli_2 appli_1', [call()])


def test_sstatus(controller, plugin, mocked_check):
    """ Test the sstatus request. """
    mocked_rpc = plugin.supvisors().get_all_process_info
    mocked_rpc.return_value = [{'application_name': 'appli_1', 'process_name': 'proc_1',
                                'statecode': 20, 'statename': 'running', 'expected_exit': True,
                                'identifiers': ['10.0.1', '10.0.2']},
                               {'application_name': 'appli_2', 'process_name': 'proc_3',
                                'statecode': 100, 'statename': 'exited', 'expected_exit': False,
                                'identifiers': []}]
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_sstatus, plugin.do_sstatus, '', [call()])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_sstatus, plugin.do_sstatus, 'all', [call()])
    # test help and request for process info from a selection of namespecs
    mocked_rpc = plugin.supvisors().get_process_info
    mocked_rpc.side_effect = [[{'application_name': 'appli_1', 'process_name': 'proc_1',
                                'statecode': 20, 'statename': 'running', 'expected_exit': True,
                                'identifiers': ['10.0.1', '10.0.2']}],
                              [{'application_name': 'appli_2', 'process_name': 'proc_3',
                                'statecode': 100, 'statename': 'exited', 'expected_exit': False,
                                'identifiers': []}]]
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_sstatus, plugin.do_sstatus,
                'appli_2:proc_3 appli_1:proc_1', [call('appli_2:proc_3'), call('appli_1:proc_1')])


def test_local_status(controller, plugin, mocked_check):
    """ Test the local_status request. """
    mocked_rpc = plugin.supvisors().get_all_local_process_info
    mocked_rpc.return_value = [{'group': 'appli_1', 'name': 'proc_1',
                                'state': 20, 'start': 1234, 'now': 4321, 'pid': 14725,
                                'extra_args': '-x dummy', 'disabled': False},
                               {'group': 'appli_2', 'name': 'proc_3',
                                'state': 0, 'start': 0, 'now': 0, 'pid': 0,
                                'extra_args': '', 'disabled': True}]
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_local_status, plugin.do_local_status,
                '', [call()])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_local_status, plugin.do_local_status,
                'all', [call()])
    # test help and request for process info from a selection of namespecs
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_local_status, plugin.do_local_status,
                'appli_2:proc_3 appli_1:proc_1', [call()])


def test_application_rules(controller, plugin, mocked_check):
    """ Test the application_rules request. """
    mocked_appli = plugin.supvisors().get_all_applications_info
    mocked_appli.return_value = [{'application_name': 'appli_1'}, {'application_name': 'appli_2'},
                                 {'application_name': 'appli_3'}]
    mocked_rpc = plugin.supvisors().get_application_rules
    returned_rules = [{'application_name': 'appli_1', 'managed': True, 'distribution': 'SINGLE_NODE',
                       'start_sequence': 2, 'stop_sequence': 3, 'starting_strategy': 'CONFIG',
                       'starting_failure_strategy': 'ABORT', 'running_failure_strategy': 'CONTINUE'},
                      {'application_name': 'appli_2', 'managed': True,
                       'distribution': 'ALL_INSTANCES', 'addresses': ['10.0.0.1', '10.0.0.2', '10.0.0.3'],
                       'start_sequence': 1, 'stop_sequence': 0, 'starting_strategy': 'LESS_LOADED',
                       'starting_failure_strategy': 'CONTINUE', 'running_failure_strategy': 'RESTART_APPLICATION'},
                      {'application_name': 'appli_3', 'managed': False}]
    # first possibility: no argument
    mocked_rpc.side_effect = returned_rules
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_application_rules, plugin.do_application_rules,
                '', [call('appli_1'), call('appli_2'), call('appli_3')])
    assert mocked_appli.call_args_list == [call(), call()]
    mocked_appli.reset_mock()
    # second possibility: use 'all'
    mocked_rpc.side_effect = returned_rules
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_application_rules, plugin.do_application_rules,
                'all', [call('appli_1'), call('appli_2'), call('appli_3')])
    assert mocked_appli.call_args_list == [call(), call()]
    mocked_appli.reset_mock()
    # test help and request for rules from a selection of application names
    mocked_rpc.side_effect = returned_rules
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_application_rules, plugin.do_application_rules,
                'appli_3 appli_2 appli_1', [call('appli_3'), call('appli_2'), call('appli_1')])
    assert mocked_appli.call_count == 0
    # test help and request with get_all_applications_info error
    mocked_appli.reset_mock()
    mocked_appli.side_effect = xmlrpclib.Fault(0, 'error')
    plugin.do_application_rules('')
    assert mocked_appli.call_args_list == [call()]
    assert not mocked_rpc.called
    _check_output_error(controller, True)


def test_process_rules(controller, plugin, mocked_check):
    """ Test the process_rules request. """
    mocked_info = plugin.supvisors().get_all_process_info
    mocked_info.return_value = [{'application_name': 'appli_1', 'process_name': 'proc_1'},
                                {'application_name': 'appli_2', 'process_name': 'proc_3'}]
    mocked_rpc = plugin.supvisors().get_process_rules
    returned_rules = [[{'application_name': 'appli_1', 'process_name': 'proc_1',
                        'identifiers': ['10.0.0.1', '10.0.0.2'],
                        'start_sequence': 2, 'stop_sequence': 3, 'required': True, 'wait_exit': False,
                        'expected_loading': 50, 'running_failure_strategy': 1}],
                      [{'application_name': 'appli_2', 'process_name': 'proc_3', 'identifiers': ['*'],
                        'start_sequence': 1, 'stop_sequence': 0, 'required': False, 'wait_exit': True,
                        'expected_loading': 15, 'running_failure_strategy': 2}]]
    # first case: no argument
    mocked_rpc.side_effect = returned_rules
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_process_rules, plugin.do_process_rules,
                '', [call('appli_1:proc_1'), call('appli_2:proc_3')])
    assert mocked_info.call_args_list == [call(), call()]
    mocked_info.reset_mock()
    # second case: use 'all'
    mocked_rpc.side_effect = returned_rules
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_process_rules, plugin.do_process_rules,
                'all', [call('appli_1:proc_1'), call('appli_2:proc_3')])
    assert mocked_info.call_args_list == [call(), call()]
    mocked_info.reset_mock()
    # test help and request for rules from a selection of namespecs
    mocked_rpc.side_effect = returned_rules
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_process_rules, plugin.do_process_rules,
                'appli_1:proc_1 appli_2:proc_3', [call('appli_1:proc_1'), call('appli_2:proc_3')])
    assert not mocked_info.called
    # test help and request with get_all_applications_info error
    mocked_info.reset_mock()
    mocked_info.side_effect = xmlrpclib.Fault(0, 'error')
    plugin.do_process_rules('')
    assert mocked_info.call_args_list == [call()]
    assert not mocked_rpc.called
    _check_output_error(controller, True)


def test_conflicts(controller, plugin, mocked_check):
    """ Test the conflicts request. """
    mocked_rpc = plugin.supvisors().get_conflicts
    mocked_rpc.return_value = [{'application_name': 'appli_1', 'process_name': 'proc_1', 'statename': 'running',
                                'identifiers': ['10.0.0.1', '10.0.0.2']},
                               {'application_name': 'appli_2', 'process_name': 'proc_3', 'statename': 'stopped',
                                'identifiers': ['10.0.0.2', '10.0.0.3']}]
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_conflicts, plugin.do_conflicts, '', [call()])


def test_start_application(controller, plugin, mocked_check):
    """ Test the start_application request. """
    mocked_appli = plugin.supvisors().get_all_applications_info
    mocked_rpc = plugin.supvisors().start_application
    _check_start_application_command(controller, mocked_check, mocked_appli, mocked_rpc,
                                     plugin.help_start_application, plugin.do_start_application,
                                     ('appli_1', 'appli_2'), 'appli_2 appli_1 appli_3 dummy_appli',
                                     ('appli_2', 'appli_1'))


def test_test_start_application(controller, plugin, mocked_check):
    """ Test the test_start_application request. """
    mocked_appli = plugin.supvisors().get_application_info
    mocked_appli.return_value = {'application_name': 'appli_1', 'managed': False}
    mocked_rpc = plugin.supvisors().test_start_application
    mocked_rpc.return_value = [{'application_name': 'appli_1', 'process_name': 'proc_1',
                                'state': 'RUNNING', 'running_identifiers': ['10.0.0.1'], 'forced_reason': ''},
                               {'application_name': 'appli_1', 'process_name': 'proc_2',
                                'state': 'EXITED', 'running_identifiers': [], 'forced_reason': ''},
                               {'application_name': 'appli_1', 'process_name': 'proc_3',
                                'state': 'FATAL', 'running_identifiers': [], 'forced_reason': 'no resource'}]
    # test the request using few arguments
    plugin.do_test_start_application('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    assert not mocked_appli.called
    mocked_check.reset_mock()
    # test the request using unknown strategy
    plugin.do_test_start_application('strategy appli_1')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    assert not mocked_appli.called
    mocked_check.reset_mock()
    # test request with unknown application
    mocked_appli.side_effect = xmlrpclib.Fault(0, 'error')
    plugin.do_test_start_application('MOST_LOADED appli_1')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    assert mocked_appli.call_args_list == [call('appli_1')]
    mocked_check.reset_mock()
    mocked_appli.reset_mock()
    # test request with application not managed
    mocked_appli.side_effect = None
    plugin.do_test_start_application('CONFIG appli_1')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    assert mocked_appli.call_args_list == [call('appli_1')]
    mocked_check.reset_mock()
    mocked_appli.reset_mock()
    # test request with application managed
    mocked_appli.return_value = {'application_name': 'appli_1', 'managed': True}
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_test_start_application,
                plugin.do_test_start_application, 'LESS_LOADED appli_1', [call(1, 'appli_1')])
    assert mocked_appli.call_args_list == [call('appli_1'), call('appli_1')]


def test_restart_application(controller, plugin, mocked_check):
    """ Test the restart_application request. """
    mocked_appli = plugin.supvisors().get_all_applications_info
    mocked_rpc = plugin.supvisors().restart_application
    _check_start_application_command(controller, mocked_check, mocked_appli, mocked_rpc,
                                     plugin.help_restart_application, plugin.do_restart_application,
                                     ('appli_1', 'appli_2'), 'appli_2 appli_1 appli_3 dummy_appli',
                                     ('appli_2', 'appli_1'))


def test_stop_application(controller, plugin, mocked_check):
    """ Test the stop_application request. """
    mocked_appli = plugin.supvisors().get_all_applications_info
    mocked_rpc = plugin.supvisors().stop_application
    _check_stop_application_command(controller, mocked_check, mocked_appli, mocked_rpc,
                                    plugin.help_stop_application, plugin.do_stop_application,
                                    ('appli_1', 'appli_2'), 'appli_2 appli_1 dummy_appli', ('appli_2', 'appli_1'))


def test_start_args(controller, plugin, mocked_check):
    """ Test the start_args request. """
    # test request to start process without extra arguments
    plugin.do_start_args('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test request to start process with extra arguments
    mocked_rpc = plugin.supvisors().start_args
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_start_args, plugin.do_start_args,
                'proc -x 5', [call('proc', '-x 5')])


def test_all_start(mocker, controller, plugin, mocked_check):
    """ Test the all_start request. """
    mocked_get_instances = mocker.patch.object(plugin, 'get_running_instances')
    mocked_rpc = Mock(**{'startProcess.return_value': True})
    mocked_get_proxy = mocker.patch.object(plugin, 'get_server_proxy', return_value=mocked_rpc)
    # test request to start process without arguments
    plugin.do_all_start('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test without running instances (doesn't really make sense)
    mocked_get_instances.return_value = {}
    plugin.do_all_start('appli_1:proc_1')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    assert not mocked_get_proxy.called
    assert not mocked_rpc.startProcess.called
    # test with running instances
    mocked_get_instances.return_value = {'node_1': 'http://10.0.0.1:30000', 'node_2': 'http://10.0.0.2:31000'}
    _check_call(controller, mocked_check, mocked_rpc.startProcess, plugin.help_all_start, plugin.do_all_start,
                'appli_1:proc_1', [call('appli_1:proc_1', False), call('appli_1:proc_1', False)])
    # get_server_proxy is called 2x2 times, without error and with error (which is set on startProcess)
    assert mocked_get_proxy.call_args_list == [call('http://10.0.0.1:30000', 'supervisor'),
                                               call('http://10.0.0.2:31000', 'supervisor'),
                                               call('http://10.0.0.1:30000', 'supervisor'),
                                               call('http://10.0.0.2:31000', 'supervisor')]


def test_all_start_args(mocker, controller, plugin, mocked_check):
    """ Test the all_start_args request. """
    mocked_get_instances = mocker.patch.object(plugin, 'get_running_instances')
    mocked_rpc = Mock(**{'start_args.return_value': True})
    mocked_get_proxy = mocker.patch.object(plugin, 'get_server_proxy', return_value=mocked_rpc)
    # test request to start process without arguments
    plugin.do_all_start_args('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test request to start process without extra arguments
    plugin.do_all_start_args('appli_1:proc_1')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test without running instances (doesn't really make sense)
    mocked_get_instances.return_value = {}
    plugin.do_all_start_args('appli_1:proc_1 -x 5')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    assert not mocked_get_proxy.called
    assert not mocked_rpc.start_args.called
    # test with running instances
    mocked_get_instances.return_value = {'node_1': 'http://10.0.0.1:30000', 'node_2': 'http://10.0.0.2:31000'}
    _check_call(controller, mocked_check, mocked_rpc.start_args, plugin.help_all_start_args, plugin.do_all_start_args,
                'appli_1:proc_1 -x 5 ', [call('appli_1:proc_1', '-x 5', False), call('appli_1:proc_1', '-x 5', False)])
    # get_server_proxy is called 2x2 times, without error and with error (which is set on startProcess)
    assert mocked_get_proxy.call_args_list == [call('http://10.0.0.1:30000', 'supvisors'),
                                               call('http://10.0.0.2:31000', 'supvisors'),
                                               call('http://10.0.0.1:30000', 'supvisors'),
                                               call('http://10.0.0.2:31000', 'supvisors')]


def test_start_process(controller, plugin, mocked_check):
    """ Test the start_process request. """
    mocked_info = plugin.supvisors().get_all_process_info
    mocked_rpc = plugin.supvisors().start_process
    _check_start_process_command(controller, mocked_check, mocked_info, mocked_rpc,
                                 plugin.help_start_process, plugin.do_start_process,
                                 ('appli_1:proc_1', 'appli_2:proc_3'), 'appli_1:proc_1 appli_2:proc_3',
                                 ('appli_1:proc_1', 'appli_2:proc_3'))


def test_test_start_process(controller, plugin, mocked_check):
    """ Test the test_start_process request. """
    mocked_rpc = plugin.supvisors().test_start_process
    mocked_rpc.return_value = [{'application_name': 'appli_1', 'process_name': 'proc_1',
                                'state': 'RUNNING', 'running_identifiers': ['10.0.0.1'], 'forced_reason': ''},
                               {'application_name': 'appli_2', 'process_name': 'proc_2',
                                'state': 'EXITED', 'running_identifiers': [], 'forced_reason': ''},
                               {'application_name': 'appli_3', 'process_name': 'proc_3',
                                'state': 'FATAL', 'running_identifiers': [], 'forced_reason': 'no resource'}]
    # test the request using few arguments
    plugin.do_test_start_process('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using unknown strategy
    plugin.do_test_start_process('strategy appli_1:proc_1')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test request with application managed
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_test_start_process,
                plugin.do_test_start_process, 'LOCAL appli_1:*', [call(3, 'appli_1:*')])


def test_start_any_process(controller, plugin, mocked_check):
    """ Test the start_any_process request. """
    # test the request using few arguments
    plugin.do_start_any_process('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    plugin.do_start_any_process('CONFIG')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using unknown strategy
    plugin.do_start_any_process('strategy regex')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test request to start the process
    mocked_rpc = plugin.supvisors().start_any_process
    _check_call(controller, mocked_check, mocked_rpc,
                plugin.help_start_any_process, plugin.do_start_any_process,
                'LESS_LOADED :x [abc]', [call(1, ':x'), call(1, '[abc]')])


def test_start_any_process_args(controller, plugin, mocked_check):
    """ Test the start_any_process_args request. """
    # test the request using few arguments
    plugin.do_start_any_process_args('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    plugin.do_start_any_process_args('CONFIG')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    plugin.do_start_any_process_args('CONFIG regex')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using unknown strategy
    plugin.do_start_any_process_args('strategy regex a list of arguments')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test request to start the process
    mocked_rpc = plugin.supvisors().start_any_process
    _check_call(controller, mocked_check, mocked_rpc,
                plugin.help_start_any_process_args, plugin.do_start_any_process_args,
                'LESS_LOADED :x  a list of arguments', [call(1, ':x', 'a list of arguments')])


def test_restart_process(controller, plugin, mocked_check):
    """ Test the restart_process request. """
    mocked_info = plugin.supvisors().get_all_process_info
    mocked_rpc = plugin.supvisors().restart_process
    _check_start_process_command(controller, mocked_check, mocked_info, mocked_rpc,
                                 plugin.help_restart_process, plugin.do_restart_process,
                                 ('appli_1:proc_1', 'appli_2:proc_3'), 'appli_1:proc_1 appli_2:proc_3',
                                 ('appli_1:proc_1', 'appli_2:proc_3'))


def test_start_process_args(controller, plugin, mocked_check):
    """ Test the start_process_args request. """
    # test the request using few arguments
    plugin.do_start_process_args('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    plugin.do_start_process_args('CONFIG')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    plugin.do_start_process_args('CONFIG proc')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using unknown strategy
    plugin.do_start_process_args('strategy program list of arguments')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test request to start the process
    mocked_rpc = plugin.supvisors().start_process
    _check_call(controller, mocked_check, mocked_rpc,
                plugin.help_start_process_args, plugin.do_start_process_args,
                'LESS_LOADED appli_2:proc_3 a list of arguments',
                [call(1, 'appli_2:proc_3', 'a list of arguments')])


def test_stop_process(controller, plugin, mocked_check):
    """ Test the stop_process request. """
    mocked_info = plugin.supvisors().get_all_process_info
    mocked_rpc = plugin.supvisors().stop_process
    _check_stop_process_command(controller, mocked_check, mocked_info, mocked_rpc,
                                plugin.help_stop_process, plugin.do_stop_process,
                                ('appli_1:proc_1', 'appli_2:proc_3'), 'appli_1:proc_1 appli_2:proc_3',
                                ('appli_1:proc_1', 'appli_2:proc_3'))


def test_update_numprocs(controller, plugin, mocked_check):
    """ Test the update_numprocs request. """
    plugin.do_update_numprocs('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using incorrect numprocs
    plugin.do_update_numprocs('dummy_process deux')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using incorrect numprocs
    plugin.do_update_numprocs('dummy_process 0')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    mocked_rpc = plugin.supvisors().update_numprocs
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_update_numprocs, plugin.do_update_numprocs,
                'dummy_process 2', [call('dummy_process', 2)])


def test_lazy_update_numprocs(controller, plugin, mocked_check):
    """ Test the lazy_update_numprocs request. """
    plugin.do_lazy_update_numprocs('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using incorrect numprocs
    plugin.do_lazy_update_numprocs('dummy_process deux')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using incorrect numprocs
    plugin.do_lazy_update_numprocs('dummy_process 0')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    mocked_rpc = plugin.supvisors().update_numprocs
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_lazy_update_numprocs, plugin.do_lazy_update_numprocs,
                'dummy_process 2', [call('dummy_process', 2, True, True)])


def test_enable(controller, plugin, mocked_check):
    """ Test the enable request. """
    plugin.do_enable('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    mocked_rpc = plugin.supvisors().enable
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable, plugin.do_enable,
                'dummy_process', [call('dummy_process')])


def test_disable(controller, plugin, mocked_check):
    """ Test the disable request. """
    plugin.do_disable('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    mocked_rpc = plugin.supvisors().disable
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable, plugin.do_disable,
                'dummy_process', [call('dummy_process')])


def test_conciliate(controller, plugin, mocked_check):
    """ Test the conciliate request. """
    plugin.do_conciliate('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using unknown strategy
    plugin.do_conciliate('strategy')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    mocked_rpc = plugin.supvisors().conciliate
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_conciliate, plugin.do_conciliate,
                'SENICIDE', [call(0)])


def test_restart_sequence(controller, plugin, mocked_check):
    """ Test the restart_sequence request. """
    mocked_rpc = plugin.supvisors().restart_sequence
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_restart_sequence, plugin.do_restart_sequence,
                '', [call()])


def test_sreload(controller, plugin, mocked_check):
    """ Test the sreload request. """
    mocked_rpc = plugin.supvisors().restart
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_sreload, plugin.do_sreload, '', [call()])


def test_sshutdown(controller, plugin, mocked_check):
    """ Test the sshutdown request. """
    mocked_rpc = plugin.supvisors().shutdown
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_sshutdown, plugin.do_sshutdown, '', [call()])


def test_end_sync(controller, plugin, mocked_check):
    """ Test the end_sync request. """
    mocked_rpc = plugin.supvisors().end_sync
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_end_sync, plugin.do_end_sync,
                '', [call('')])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_end_sync, plugin.do_end_sync,
                '10.0.0.1', [call('10.0.0.1')])
    # test error
    plugin.do_end_sync('10.0.0.1 10.0.0.2')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]


def test_loglevel(controller, plugin, mocked_check):
    """ Test the loglevel request. """
    mocked_rpc = plugin.supvisors().change_log_level
    # test the request without parameter
    plugin.do_loglevel('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using unknown strategy
    plugin.do_loglevel('logger')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    for code, level in RPCInterface.get_logger_levels().items():
        _check_call(controller, mocked_check, mocked_rpc, plugin.help_loglevel, plugin.do_loglevel,
                    level, [call(code)])


def test_enable_stats_host(controller, plugin, mocked_check):
    """ Test the enable_stats request for host. """
    mocked_rpc = plugin.supvisors().enable_host_statistics
    # test the request using unknown statistics type
    plugin.do_enable_stats('node')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable_stats, plugin.do_enable_stats,
                'host', [call(True)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable_stats, plugin.do_enable_stats,
                'host process', [call(True)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable_stats, plugin.do_enable_stats,
                'all', [call(True)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable_stats, plugin.do_enable_stats,
                '', [call(True)])


def test_enable_stats_process(controller, plugin, mocked_check):
    """ Test the enable_stats request for process. """
    mocked_rpc = plugin.supvisors().enable_process_statistics
    # test the request using unknown statistics type
    plugin.do_enable_stats('node')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable_stats, plugin.do_enable_stats,
                'process', [call(True)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable_stats, plugin.do_enable_stats,
                'host process', [call(True)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable_stats, plugin.do_enable_stats,
                'all', [call(True)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_enable_stats, plugin.do_enable_stats,
                '', [call(True)])


def test_disable_stats_host(controller, plugin, mocked_check):
    """ Test the disable_stats request for host. """
    mocked_rpc = plugin.supvisors().enable_host_statistics
    # test the request using unknown statistics type
    plugin.do_disable_stats('node')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable_stats, plugin.do_disable_stats,
                'host', [call(False)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable_stats, plugin.do_disable_stats,
                'host process', [call(False)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable_stats, plugin.do_disable_stats,
                'all', [call(False)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable_stats, plugin.do_disable_stats,
                '', [call(False)])


def test_disable_stats_process(controller, plugin, mocked_check):
    """ Test the disable_stats request for process. """
    mocked_rpc = plugin.supvisors().enable_process_statistics
    # test the request using unknown statistics type
    plugin.do_disable_stats('node')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable_stats, plugin.do_disable_stats,
                'process', [call(False)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable_stats, plugin.do_disable_stats,
                'host process', [call(False)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable_stats, plugin.do_disable_stats,
                'all', [call(False)])
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_disable_stats, plugin.do_disable_stats,
                '', [call(False)])


def test_stats_period(controller, plugin, mocked_check):
    """ Test the stats_period request. """
    mocked_rpc = plugin.supvisors().update_collecting_period
    # test the request without parameter
    plugin.do_stats_period('')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test the request using an invalid float
    plugin.do_stats_period('seven')
    _check_output_error(controller, True)
    assert mocked_check.call_args_list == [call()]
    mocked_check.reset_mock()
    # test help and request
    _check_call(controller, mocked_check, mocked_rpc, plugin.help_stats_period, plugin.do_stats_period,
                '7.5', [call(7.5)])


def test_upcheck(controller, plugin):
    """ Test the _upcheck method. """
    # test different API versions
    mocked_rpc = plugin.supvisors().get_api_version
    mocked_rpc.return_value = 'dummy_version'
    assert not plugin._upcheck()
    _check_output_error(controller, True)
    assert mocked_rpc.call_args_list == [call()]
    mocked_rpc.reset_mock()
    # test handled RPC error
    mocked_rpc.side_effect = xmlrpclib.Fault(Faults.UNKNOWN_METHOD, '')
    assert not plugin._upcheck()
    _check_output_error(controller, True)
    assert mocked_rpc.call_args_list == [call()]
    mocked_rpc.reset_mock()
    # test not handled RPC error
    mocked_rpc.side_effect = xmlrpclib.Fault(0, 'error')
    with pytest.raises(xmlrpclib.Fault):
        plugin._upcheck()
    assert mocked_rpc.call_args_list == [call()]
    mocked_rpc.reset_mock()
    # test handled socket errors
    mocked_rpc.side_effect = socket.error(errno.ECONNREFUSED)
    assert not plugin._upcheck()
    _check_output_error(controller, True)
    assert mocked_rpc.call_args_list == [call()]
    mocked_rpc.reset_mock()
    mocked_rpc.side_effect = socket.error(errno.ENOENT)
    assert not plugin._upcheck()
    _check_output_error(controller, True)
    assert mocked_rpc.call_args_list == [call()]
    mocked_rpc.reset_mock()
    # test not handled socket error
    mocked_rpc.side_effect = socket.error(errno.EWOULDBLOCK)
    with pytest.raises(socket.error):
        plugin._upcheck()
    assert mocked_rpc.call_args_list == [call()]
    mocked_rpc.reset_mock()
    # test normal behaviour
    mocked_rpc.side_effect = None
    mocked_rpc.return_value = __version__
    assert plugin._upcheck()
    assert mocked_rpc.call_args_list == [call()]


def test_make_plugin(mocker, controller):
    """ Test the plugin factory. """
    mocked_plugin = mocker.patch('supvisors.supvisorsctl.ControllerPlugin')
    make_supvisors_controller_plugin(controller)
    assert mocked_plugin.call_args_list == [call(controller)]


def test_main(mocker):
    """ Test the plugin factory. """
    mocked_client_options = Mock(args=['start', 'program'], interactive=False, plugin_factories=[])
    mocked_controller = Mock(exitstatus=2)
    mocker.patch('supvisors.supvisorsctl.ClientOptions', return_value=mocked_client_options)
    mocker.patch('supvisors.supvisorsctl.Controller', return_value=mocked_controller)
    mocked_sys = mocker.patch('supvisors.supvisorsctl.sys')
    # test with arguments
    main(args='command args')
    assert mocked_client_options.realize.call_args_list == [call('command args', doc=supervisorctl.__doc__)]
    assert mocked_controller.onecmd.call_args_list == [call('start program')]
    assert not mocked_controller.exec_cmdloop.called
    assert mocked_sys.exit.call_args_list == [call(2)]
    mocker.resetall()
    # test without arguments
    mocked_client_options.args = None
    mocked_client_options.interactive = True
    main()
    assert mocked_client_options.realize.call_args_list == [call(None, doc=supervisorctl.__doc__)]
    assert not mocked_controller.onecmd.called
    assert mocked_controller.exec_cmdloop.call_args_list == [call(None, mocked_client_options)]
    assert mocked_sys.exit.call_args_list == [call(0)]



# ======================================================================
# Copyright 2016 Julien LE CLEACH
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import time
from copy import copy
from typing import Any, Dict, Tuple

from supervisor.loggers import Logger
from supervisor.xmlrpc import capped_int

from .internal_com.mapper import SupvisorsInstanceId
from .process import ProcessStatus
from .ttypes import SupvisorsInstanceStates, SupvisorsStates, InvalidTransition, Payload
from .utils import TICK_PERIOD


class StateModes:

    def __init__(self, state: SupvisorsStates = SupvisorsStates.OFF,
                 discovery_mode: bool = False,
                 master_identifier: str = '',
                 starting_jobs: bool = False, stopping_jobs: bool = False):
        """ Initialization of the attributes.

        :param state: the FSM state.
        :param master_identifier: the identifier of the Supvisors Master instance.
        :param starting_jobs: the Starter progress.
        :param stopping_jobs: the Stopper progress.
        """
        self.state: SupvisorsStates = state
        self.discovery_mode: bool = discovery_mode
        self.master_identifier: str = master_identifier
        self.starting_jobs: bool = starting_jobs
        self.stopping_jobs: bool = stopping_jobs

    def __copy__(self):
        """ Create a new StateModes object with the same attributes.

        :return: a copy of this StateModes object
        """
        return type(self)(self.state, self.discovery_mode, self.master_identifier,
                          self.starting_jobs, self.stopping_jobs)

    def __eq__(self, other) -> bool:
        """ Check if the other object is equivalent.

        :param other: a StateModes object
        :return:
        """
        if isinstance(other, StateModes):
            return (self.state == other.state
                    and self.discovery_mode == other.discovery_mode
                    and self.master_identifier == other.master_identifier
                    and self.starting_jobs == other.starting_jobs
                    and self.stopping_jobs == other.stopping_jobs)
        return False

    def apply(self, fsm_state: SupvisorsStates = None, master_identifier: str = None,
              starter: bool = None, stopper: bool = None):
        """ Get the Supvisors instance state and modes with changes applied.

        :param fsm_state: the new FSM state.
        :param master_identifier: the identifier of the Master Supvisors instance.
        :param starter: the Starter progress.
        :param stopper: the Stopper progress.
        :return: True if state and modes have changed, and the changed state and modes.
        """
        if fsm_state is not None:
            self.state = fsm_state
        if master_identifier is not None:
            self.master_identifier = master_identifier
        if starter is not None:
            self.starting_jobs = starter
        if stopper is not None:
            self.stopping_jobs = stopper

    def update(self, payload: Payload) -> None:
        """ Get the Supvisors instance state and modes with changes applied.

        :param payload: the Supvisors instance state and modes
        :return: None
        """
        self.state = SupvisorsStates(payload['fsm_statecode'])
        self.discovery_mode = payload['discovery_mode']
        self.master_identifier = payload['master_identifier']
        self.starting_jobs = payload['starting_jobs']
        self.stopping_jobs = payload['stopping_jobs']

    def serial(self):
        """ Return a serializable form of the StatesModes. """
        return {'fsm_statecode': self.state.value, 'fsm_statename': self.state.name,
                'discovery_mode': self.discovery_mode,
                'master_identifier': self.master_identifier,
                'starting_jobs': self.starting_jobs,
                'stopping_jobs': self.stopping_jobs}


class SupvisorsTimes:

    def __init__(self, identifier: str, logger: Logger):
        """ Storage of counter, clock and time received from the remote Supvisors instance.
        Upon reception of a remote TICK, the correspondence is made with the local counter, clock and time.

        All calculations must be performed using the monotonic clock times, and in the local Supvisors reference.
        Times are only used for display on the Web UI.

        Attributes:
            - identifier: the identifier of the remote Supvisors instance ;
            - remote_sequence_counter: the TICK counter ;
            - remote_mtime: the last monotonic time received from the remote Supvisors instance ;
            - remote_time: the last time received from the remote Supvisors instance ;
            - local_sequence_counter: remote_sequence_counter in local counter reference ;
            - local_monotonic: remote_monotonic in local monotonic time reference ;
            - local_mtime: remote_time in local time reference ;
            - start_local_mtime: the approximate start time of the remote Supvisors instance,
                in local monotonic time reference.
        """
        self.identifier: str = identifier
        self.logger: Logger = logger
        # counter and times of the remote Supvisors instance
        self.remote_sequence_counter: int = 0
        self.remote_mtime: float = 0.0
        self.remote_time: float = 0.0
        # corresponding counter and times of the local Supvisors instance
        self.local_sequence_counter: int = 0
        self.local_mtime: float = 0.0
        self.local_time: float = 0.0
        # approximate startup monotonic time of the remote Supvisors instance (in the local monotonic time reference)
        # will be used to display the remote Supvisors instance uptime
        self.start_local_mtime: float = -1.0

    @property
    def capped_remote_time(self) -> int:
        """ Return a remote time compliant with XML-RPC limits.
        The xml-rpc integer type will be saturated in jan 2038 ("year 2038 problem").
        """
        return capped_int(self.remote_time)

    @property
    def capped_local_time(self) -> int:
        """ Return a local time compliant with XML-RPC limits.
        The xml-rpc integer type will be saturated in jan 2038 ("year 2038 problem").
        """
        return capped_int(self.local_time)

    def get_current_uptime(self) -> float:
        """ Return the duration since the approximate stat time of the remote Supvisors instance. """
        return time.monotonic() - self.start_local_mtime

    def get_current_remote_time(self, local_mtime: float) -> float:
        """ Return the current time of the remote Supvisors instance.

        Add the last remote time received to the monotonic duration from the last clock time stored.

        :param local_mtime: the monotonic timestamp reference.
        :return: the remote time.
        """
        return self.remote_time + (local_mtime - self.local_mtime)

    def update(self, remote_sequence_counter: int, remote_mtime: float, remote_time: float,
               local_sequence_counter: int):
        """ Update the time counters of the remote Supvisors instance, and make correspondence with self time.

        :param remote_sequence_counter: the TICK counter received from the remote Supvisors instance.
        :param remote_mtime: the monotonic timestamp received from the remote Supvisors instance.
        :param remote_time: the timestamp received from the remote Supvisors instance.
        :param local_sequence_counter: the latest sequence counter received from the local Supvisors instance.
        :return: None.
        """
        if local_sequence_counter < 0:
            # remote Supvisors instance is local Supvisors instance, so use the same input data
            local_sequence_counter = remote_sequence_counter
            local_mtime = remote_mtime
            local_time = remote_time
        else:
            local_mtime = time.monotonic()
            local_time = time.time()
        # check sequence counter to identify stealth supervisor restart
        # (only for remote, cannot happen with local)
        if remote_sequence_counter < self.remote_sequence_counter:
            self.logger.warn(f'SupvisorsTimes.update: stealth restart of Supvisors={self.identifier}')
            # Force Supvisors inactivity by resetting its local_sequence_counter
            # The Supvisors periodical check will handle the node invalidation
            local_sequence_counter = 0
            self.start_local_mtime = -1
        # update remote attributes
        self.remote_sequence_counter = remote_sequence_counter
        self.remote_mtime = remote_mtime
        self.remote_time = remote_time
        # update local correspondent attributes
        self.local_sequence_counter = local_sequence_counter
        self.local_mtime = local_mtime
        self.local_time = local_time
        # approximation of the remote Supvisors instance startup clock (in local clock reference)
        #     from remote_sequence_counter and TICK_PERIOD
        #     good enough as it is just for Web UI display
        if self.start_local_mtime < 0:
            self.start_local_mtime = local_mtime - TICK_PERIOD * remote_sequence_counter

    def serial(self):
        """ Return a serializable form of the SupvisorsTimes instance. """
        return {'remote_sequence_counter': self.remote_sequence_counter,
                'remote_time': self.capped_remote_time,
                'remote_mtime': self.remote_mtime,
                'local_sequence_counter': self.local_sequence_counter,
                'local_time': self.capped_local_time,
                'local_mtime': self.local_mtime}


class SupvisorsInstanceStatus:
    """ Class defining the status of a Supvisors instance.

    Attributes:
        - supvisors_id: the parameters identifying where the Supvisors instance is expected to be running ;
        - state: the state of the Supvisors instance in SupvisorsInstanceStates ;
        - time: the counter, time and clock of the remote Supvisors instance associated to the local ones ;
        - processes: the list of processes that are configured in the Supervisor of the Supvisors instance ;
        - state_modes: the Supvisors instance state and modes.
    """

    def __init__(self, supvisors_id: SupvisorsInstanceId, supvisors: Any):
        """ Initialization of the attributes. """
        self.supvisors = supvisors
        # attributes
        self.supvisors_id: SupvisorsInstanceId = supvisors_id
        self._state: SupvisorsInstanceStates = SupvisorsInstanceStates.UNKNOWN
        self.times: SupvisorsTimes = SupvisorsTimes(self.identifier, self.logger)
        self.processes: Dict[str, ProcessStatus] = {}
        # state and modes
        self.state_modes = StateModes()
        # the local instance may use the process statistics collector
        self.stats_collector = None
        is_local = supvisors.mapper.local_identifier == self.identifier
        if is_local:
            # use the condition to set the local discovery mode in states / modes object
            self.state_modes.discovery_mode = supvisors.options.discovery_mode
            # copy the process collector reference
            if supvisors.options.process_stats_enabled:
                self.stats_collector = supvisors.stats_collector

    @property
    def logger(self) -> Logger:
        """ Shortcut to the Supvisors logger. """
        return self.supvisors.logger

    def reset(self):
        """ Reset the contextual part of the Supvisors instance.
        Silent and isolated Supvisors instances are not reset.

        :return: None.
        """
        if self.has_active_state():
            # do NOT use state setter as transition may be rejected
            self._state = SupvisorsInstanceStates.UNKNOWN
        # use a new SupvisorsTimes instance
        self.times = SupvisorsTimes(self.identifier, self.logger)

    # accessors / mutators
    @property
    def identifier(self):
        """ Property getter for the 'identifier' attribute of the Supvisors instance. """
        return self.supvisors_id.identifier

    @property
    def usage_identifier(self):
        """ Property getter for the representation of the Supvisors instance. """
        return str(self.supvisors_id)

    @property
    def state(self) -> SupvisorsInstanceStates:
        """ Property getter for the 'state' attribute. """
        return self._state

    @state.setter
    def state(self, new_state: SupvisorsInstanceStates):
        """ Property setter for the 'state' attribute. """
        if self._state != new_state:
            if not self.check_transition(new_state):
                raise InvalidTransition(f'SupvisorsInstanceStatus.state: Supvisors={self.usage_identifier}'
                                        f' transition rejected from {self.state.name} to {new_state.name}')
            self._state = new_state
            self.logger.warn(f'SupvisorsInstanceStatus.state: Supvisors={self.usage_identifier} is {self.state.name}')
            # TODO: export status ? from context
            if new_state in [SupvisorsInstanceStates.SILENT, SupvisorsInstanceStates.ISOLATED]:
                self.logger.debug(f'SupvisorsInstanceStatus.state: FSM is OFF in Supvisors={self.usage_identifier}')
                self.state_modes.state = SupvisorsStates.OFF
                # TODO: export states and modes ?

    @property
    def isolated(self) -> bool:
        """ Return True if the Supvisors instance is isolated. """
        return self.state == SupvisorsInstanceStates.ISOLATED

    @property
    def sequence_counter(self):
        """ the remote sequence counter will be used as a reference outside of this class. """
        return self.times.remote_sequence_counter

    # serialization
    def serial(self) -> Payload:
        """ Return a serializable form of the SupvisorsInstanceStatus. """
        payload = {'identifier': self.identifier,
                   'nick_identifier': self.supvisors_id.nick_identifier,
                   'node_name': self.supvisors_id.host_id,
                   'port': self.supvisors_id.http_port,
                   'statecode': self.state.value, 'statename': self.state.name,
                   'loading': self.get_load(),
                   'process_failure': self.has_error()}
        payload.update(self.times.serial())
        payload.update(self.state_modes.serial())
        return payload

    # methods
    def update_state_modes(self, event: Payload) -> None:
        """ Update the Supvisors instance state and modes.

        :param event: the state or the mode updated.
        :return: None.
        """
        self.state_modes.update(event)

    def apply_state_modes(self, event: Payload) -> Tuple[bool, StateModes]:
        """ Apply the change on a copy of states_modes.

        :param event: the state or the mode updated.
        :return: None.
        """
        ref_state_modes = copy(self.state_modes)
        self.state_modes.apply(**event)
        return ref_state_modes != self.state_modes, self.state_modes

    def has_active_state(self) -> bool:
        """ Return True if the instance status is in an active state.

        :return: the activity status.
        """
        return self.state in [SupvisorsInstanceStates.CHECKING, SupvisorsInstanceStates.CHECKED,
                              SupvisorsInstanceStates.RUNNING]

    def is_inactive(self, local_sequence_counter: int) -> bool:
        """ Return True if the latest update was received more than INACTIVITY_TICKS ago.

        :param local_sequence_counter: the current local sequence counter.
        :return: the inactivity status.
        """
        # NOTE: by design, there will be always a gap of 1 between the local_sequence_counter and the latest
        #       local_sequence_counter of the remote Supvisors instance, unless the remote Supvisors instance is stopped
        #       or unreachable, because the periodic check is performed on the new local TICK.
        counter_diff = local_sequence_counter - self.times.local_sequence_counter
        return self.has_active_state() and counter_diff > self.supvisors.options.inactivity_ticks

    def update_tick(self, remote_sequence_counter: int, remote_mtime: float, remote_time: float,
                    local_sequence_counter: int = -1):
        """ Update the time attributes of the current object, including the time attributes of all its processes.

        :param remote_sequence_counter: the TICK counter received from the Supvisors instance ;
        :param remote_time: the timestamp received from the Supvisors instance ;
        :param remote_mtime: the timestamp received from the Supvisors instance ;
        :param local_sequence_counter: the last TICK counter received from the local Supvisors instance ;
        :return: None
        """
        self.logger.debug(f'SupvisorsInstanceStatus.update_tick: update Supvisors={self.usage_identifier}' 
                          f' with sequence_counter={remote_sequence_counter} remote_time={remote_time}'
                          f' remote_mtime={remote_mtime} local_sequence_counter={local_sequence_counter}')
        self.times.update(remote_sequence_counter, remote_mtime, remote_time, local_sequence_counter)
        # update all process times
        for process in self.processes.values():
            process.update_times(self.identifier, self.times.remote_mtime, self.times.remote_time)

    def check_transition(self, new_state):
        """ Check that the state transition is valid. """
        return new_state in self._Transitions[self.state]

    # methods on processes
    def add_process(self, process: ProcessStatus) -> None:
        """ Add a new process to the process list.

        :param process: the process status to be added to the Supvisors instance
        :return: None
        """
        self.processes[process.namespec] = process
        # update the collector withe process if it is already running
        if self.stats_collector:
            pid = process.get_pid(self.identifier)
            if pid > 0:
                self.stats_collector.send_pid(process.namespec, pid)

    def update_process(self, process: ProcessStatus) -> None:
        """ Upon a process state change, check if a pid is available to update the collector.

        :param process: the process status that has been updated
        :return: None
        """
        if self.stats_collector:
            pid = process.get_pid(self.identifier)
            self.stats_collector.send_pid(process.namespec, pid)

    def remove_process(self, process: ProcessStatus) -> None:
        """ Remove a process from the process list.

        :param process: the process to be removed from the Supvisors instance
        :return: None
        """
        del self.processes[process.namespec]
        # update the collector
        if self.stats_collector:
            self.stats_collector.send_pid(process.namespec, 0)

    def running_processes(self):
        """ Return the process running on the Supvisors instance.
        Here, 'running' means that the process state is in Supervisor RUNNING_STATES. """
        return [process for process in self.processes.values()
                if process.running_on(self.identifier)]

    def get_load(self) -> int:
        """ Return the load of the Supvisors instance, by summing the declared load of the processes running
        on the Supvisors instance.

        :return: the total load
        """
        instance_load = sum(process.rules.expected_load for process in self.running_processes())
        self.logger.trace(f'SupvisorsInstanceStatus.get_load: Supvisors={self.usage_identifier} load={instance_load}')
        return instance_load

    def has_error(self) -> bool:
        """ Return True if any process managed by the local Supervisor is in failure.

        :return: the error status
        """
        return (self.state == SupvisorsInstanceStates.RUNNING
                and any(process.crashed(self.identifier)
                        for process in self.processes.values()))

    # dictionary for transitions
    _Transitions = {SupvisorsInstanceStates.UNKNOWN: (SupvisorsInstanceStates.CHECKING,
                                                      SupvisorsInstanceStates.ISOLATED,
                                                      SupvisorsInstanceStates.SILENT),
                    SupvisorsInstanceStates.CHECKING: (SupvisorsInstanceStates.UNKNOWN,
                                                       SupvisorsInstanceStates.CHECKED,
                                                       SupvisorsInstanceStates.ISOLATED,
                                                       SupvisorsInstanceStates.SILENT),
                    SupvisorsInstanceStates.CHECKED: (SupvisorsInstanceStates.RUNNING,
                                                      SupvisorsInstanceStates.ISOLATED,
                                                      SupvisorsInstanceStates.SILENT),
                    SupvisorsInstanceStates.RUNNING: (SupvisorsInstanceStates.SILENT,
                                                      SupvisorsInstanceStates.ISOLATED,
                                                      SupvisorsInstanceStates.CHECKING),
                    SupvisorsInstanceStates.SILENT: (SupvisorsInstanceStates.CHECKING,
                                                     SupvisorsInstanceStates.ISOLATED),
                    SupvisorsInstanceStates.ISOLATED: ()
                    }



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import os
import socket
import time
from typing import Any, Dict, List, Optional

from supervisor.datatypes import Syslog
from supervisor.events import notify
from supervisor.http import supervisor_auth_handler, logtail_handler, mainlogtail_handler
from supervisor.loggers import Logger
from supervisor.medusa import default_handler, filesys
from supervisor.options import make_namespec, split_namespec, ProcessConfig
from supervisor.process import ProcessGroupBase, Subprocess
from supervisor.states import ProcessStates, STOPPED_STATES

from .ttypes import (ProcessAddedEvent, ProcessRemovedEvent, ProcessEnabledEvent, ProcessDisabledEvent,
                     SupvisorsProcessConfig, GroupConfigInfo, NameList)

SUPERVISOR_TAIL_DEFAULT = 1024


def spawn(self):
    """ Overridden Subprocess spawn to handle disabled and obsolete processes.

    :return: the process id or None if the fork() call fails.
    """
    if not self.supvisors_config.program_config.disabled and not self.obsolete:
        return self._spawn()


class LogtailHandler(logtail_handler):
    """ Specialization of Supervisor logtail_handler to enable tail size configuration. """

    def handle_request(self, request):
        # call parent
        super().handle_request(request)
        # get configured tail size
        new_head = self.supervisord.supvisors.options.tailf_limit
        # re-calculate sz
        sz = request.outgoing[-1].sz
        request.outgoing[-1].sz = min(sz, max(sz + SUPERVISOR_TAIL_DEFAULT - new_head, 0))


class MainLogtailHandler(mainlogtail_handler):
    """ Specialization of Supervisor mainlogtail_handler to enable tail size configuration. """

    def handle_request(self, request):
        # call parent
        super().handle_request(request)
        # get configured tail size
        new_head = self.supervisord.supvisors.options.tailf_limit
        # re-calculate sz
        sz = request.outgoing[-1].sz
        request.outgoing[-1].sz = min(sz, max(sz + SUPERVISOR_TAIL_DEFAULT - new_head, 0))


class SupervisorData:
    """ Supvisors is started in Supervisor so Supervisor internal data is available from the supervisord structure. """

    # shortcuts
    _system_rpc_interface = None
    _supervisor_rpc_interface = None
    _supvisors_rpc_interface = None

    def __init__(self, supvisors, supervisord):
        """ Initialization of the attributes.

        :param supvisors: the Supvisors global structure
        :param supervisord: the Supervisor global structure
        """
        self.supvisors = supvisors
        self.supervisord = supervisord
        # check HTTP configuration
        for config in supervisord.options.server_configs:
            if config['family'] == socket.AF_INET:
                self.server_config = config
                break
        else:
            # there MUST be an inet HTTP server
            raise ValueError(f'Supervisor MUST be configured using inet_http_server: {supervisord.options.configfile}')

    @property
    def logger(self) -> Logger:
        """ Get the Supvisors logger. """
        return self.supvisors.logger

    @property
    def http_server(self):
        """ Get the internal Supervisor HTTP server structure.

        :return: the HTTP server structure
        """
        for config, hs in self.supervisord.options.httpservers:
            if config['family'] == socket.AF_INET:
                return hs

    @property
    def system_rpc_interface(self):
        """ Get the internal System Supervisor RPC handler.
        XML-RPC call in another XML-RPC call on the same server is blocking.

        :return: the System Supervisor RPC handler
        """
        if not self._system_rpc_interface:
            # the first handler is the XML-RPC interface for rapid access
            handler = self.http_server.handlers[0]
            # if authentication used, handler is wrapped
            if self.username:
                handler = handler.handler
            self._system_rpc_interface = handler.rpcinterface.system
        return self._system_rpc_interface

    @property
    def supervisor_rpc_interface(self):
        """ Get the internal Supervisor RPC handler.
        XML-RPC call in another XML-RPC call on the same server is blocking.

        :return: the Supervisor RPC handler
        """
        if not self._supervisor_rpc_interface:
            # the first handler is the XML-RPC interface for rapid access
            handler = self.http_server.handlers[0]
            # if authentication used, handler is wrapped
            if self.username:
                handler = handler.handler
            self._supervisor_rpc_interface = handler.rpcinterface.supervisor
        return self._supervisor_rpc_interface

    @property
    def supvisors_rpc_interface(self):
        """ Get the internal Supvisors RPC handler.
        XML-RPC call in another XML-RPC call on the same server is blocking.

        :return: the Supvisors RPC handler
        """
        if not self._supvisors_rpc_interface:
            # the first handler is the XML-RPC interface for rapid access
            handler = self.http_server.handlers[0]
            # if authentication is used, handler is wrapped
            if self.username:
                handler = handler.handler
            self._supvisors_rpc_interface = handler.rpcinterface.supvisors
        return self._supvisors_rpc_interface

    @property
    def identifier(self) -> str:
        """ Get the internal Supervisor identifier.

        :return: the Supervisor identifier
        """
        return self.supervisord.options.identifier

    @property
    def server_host(self):
        """ Return the host defined in the inet http server section. """
        return self.server_config['host'] or 'localhost'

    @property
    def server_port(self):
        """ Return the port defined in the inet http server section. """
        return self.server_config['port']

    @property
    def username(self) -> str:
        """ Return the user authentication defined in the inet http server section. """
        return self.server_config['username']

    @property
    def password(self) -> str:
        """ Return the password authentication defined in the inet http server section. """
        return self.server_config['password']

    @property
    def server_url(self) -> str:
        """ Return the server URL defined in the inet http server section. """
        # do NOT use Supervisor serverurl as priority is given to the unix server if defined
        return f'http://{self.server_host}:{self.server_port}'

    @property
    def supervisor_state(self):
        """ Return the supervisord internal state. """
        return self.supervisord.options.mood

    def get_env(self) -> Dict[str, str]:
        """ Return a simple environment that can be used for the configuration of the XML-RPC client. """
        return {'SUPERVISOR_SERVER_URL': self.server_url,
                'SUPERVISOR_USERNAME': self.username,
                'SUPERVISOR_PASSWORD': self.password}
    
    def replace_tail_handlers(self) -> None:
        """ This method replaces Supervisor web UI tail handlers. """
        tail_handler = LogtailHandler(self.supervisord)
        main_tail_handler = MainLogtailHandler(self.supervisord)
        # deal with authentication
        if self.username:
            # wrap the default handler in an authentication handler
            users = {self.username: self.password}
            tail_handler = supervisor_auth_handler(users, tail_handler)
            main_tail_handler = supervisor_auth_handler(users, main_tail_handler)
        else:
            self.logger.debug('SupervisorData.replace_tail_handlers: Server running without any HTTP'
                              ' authentication checking')
        # replace Supervisor handlers considering the order in supervisor.http.make_http_servers
        self.http_server.handlers[1] = tail_handler
        self.http_server.handlers[2] = main_tail_handler

    def replace_default_handler(self) -> None:
        """ This method replaces Supervisor web UI with Supvisors web UI. """
        # create default handler pointing on Supvisors ui directory
        here = os.path.abspath(os.path.dirname(__file__))
        template_dir = os.path.join(here, 'ui')
        filesystem = filesys.os_filesystem(template_dir)
        def_handler = default_handler.default_handler(filesystem)
        # deal with authentication
        if self.username:
            # wrap the default handler in an authentication handler
            users = {self.username: self.password}
            def_handler = supervisor_auth_handler(users, def_handler)
        else:
            self.logger.debug('SupervisorData.replace_default_handler: Server running without any HTTP'
                              ' authentication checking')
        # replace Supervisor default handler at the end of the list
        self.http_server.handlers[-1] = def_handler

    def close_httpservers(self) -> None:
        """ Call the close_httpservers of Supervisor.
        This is called when receiving the Supervisor stopping event in order to force the termination
        of any asynchronous job. """
        # Supervisor issue #1596:
        #     In the event of a reload, the HTTP socket will be re-opened very quickly.
        #     Despite the REUSEADDR is set, there may still be a handle somewhere on the socket that makes it
        #     NOT deallocated, although it has been closed.
        #     To avoid issues, it is better to shut the socket down before closing it.
        self.http_server.socket.shutdown(socket.SHUT_RDWR)
        self.supervisord.options.close_httpservers()
        self.supervisord.options.httpservers = ()

    # Access to Group / Process structures and configurations
    def complete_internal_data(self, process_configs: Dict[str, SupvisorsProcessConfig],
                               group_name: str = None) -> None:
        """ Add extra attributes to Supervisor internal data.

        This is triggered in two occasions:
        1) when Supervisor sends its RUNNING state ;
        2) when groups are added to Supervisor.

        Note that the initial group added events are sent before the Supvisors RPC interface is created.

        :param process_configs: the process configuration extension.
        :param group_name: the name of the group that has been added to the Supervisor internal data.
        :return: None
        """
        # determine targets
        if group_name:
            groups = [self.supervisord.process_groups[group_name]]
        else:
            groups = self.supervisord.process_groups.values()
        # apply the new configuration
        for group in groups:
            for process in group.processes.values():
                alt_process_config = process_configs[process.config.name]
                self.complete_process(process, alt_process_config)

    def get_group_processes(self, application_name: str) -> List[Subprocess]:
        """ This method returns the processes related to a group.

        :param application_name: the group name
        :return: a list of Supervisor processes
        """
        # WARN: the method may throw a KeyError exception
        return self.supervisord.process_groups[application_name].processes

    def _get_process(self, namespec: str) -> Subprocess:
        """ This method returns the process configuration related to a namespec. """
        # WARN: the method may throw a KeyError exception
        application_name, process_name = split_namespec(namespec)
        return self.get_group_processes(application_name)[process_name]

    def _get_process_config(self, namespec: str) -> ProcessConfig:
        """ This method returns the process configuration related to a namespec. """
        return self._get_process(namespec).config

    def autorestart(self, namespec: str) -> bool:
        """ This method checks if autorestart is configured on the program. """
        return self._get_process_config(namespec).autorestart is not False

    def disable_autorestart(self, namespec: str) -> None:
        """ This method forces the autorestart to False in Supervisor internal data. """
        self._get_process_config(namespec).autorestart = False

    def get_process_info(self, namespec: str) -> Dict[str, Any]:
        """ Get the process monotonic values and extra arguments.

        :param namespec: the program namespec.
        :return: a dictionary of configuration values.
        """
        process = self._get_process(namespec)
        return {'start_monotonic': process.laststart_monotonic,
                'stop_monotonic': process.laststop_monotonic,
                'now_monotonic': time.monotonic(),
                'extra_args': process.extra_args,
                'startsecs': process.config.startsecs,
                'stopwaitsecs': process.config.stopwaitsecs,
                'process_index': process.supvisors_config.process_index,
                'program_name': process.supvisors_config.program_config.name,
                'disabled': process.supvisors_config.program_config.disabled,
                'has_stdout': process.config.stdout_logfile not in [None, Syslog],
                'has_stderr': process.config.stderr_logfile not in [None, Syslog]}

    def has_logfile(self, namespec: str, channel: str) -> bool:
        """ Return True if the process has a logfile configuration on the channel.

        :param namespec: the program namespec.
        :param channel: the logfile channel (stdout or stderr).
        :return: True if the process has a logfile configured on the channel.
        """
        process_config = self._get_process_config(namespec)
        return getattr(process_config, '%s_logfile' % channel)

    # Supervisor issue #1023
    def update_extra_args(self, namespec: str, extra_args: str) -> None:
        """ This method is used to add extra arguments to the command line.
        Implementation of Supervisor issue #1023 - Pass arguments to program when starting a job.

        :param namespec: the process namespec.
        :param extra_args: the arguments to be added to the program command line.
        :return: None.
        """
        process = self._get_process(namespec)
        # reset command line
        process.config.command = process.supvisors_config.command_ref
        process.extra_args = extra_args
        # apply args to command line
        if extra_args:
            process.config.command += ' ' + extra_args
        self.logger.trace(f'SupervisorData.update_extra_args: {namespec} extra_args={extra_args}')

    def update_start(self, namespec: str) -> None:
        """ Add the monotonic start time to the internal process.

        :param namespec: the process namespec.
        :return: None.
        """
        process = self._get_process(namespec)
        process.laststart_monotonic = time.monotonic()

    def update_stop(self, namespec: str) -> None:
        """ Add the monotonic stop time to the internal process.

        :param namespec: the process namespec.
        :return: None.
        """
        # get Supervisor process from namespec
        group_name, process_name = split_namespec(namespec)
        group = self.supervisord.process_groups[group_name]
        process = group.processes[process_name]
        if process.obsolete:
            # if process is marked as obsolete following a decrease of numprocs, delete it
            self.delete_process(group, process)
        else:
            # update monotonic stop time
            process.laststop_monotonic = time.monotonic()

    def force_process_fatal(self, namespec: str, reason: str) -> None:
        """ This method forces the FATAL process state into Supervisor internal data and dispatches process event
        to event listeners. """
        process = self._get_process(namespec)
        # need to force BACKOFF state to go through assertion
        process.state = ProcessStates.BACKOFF
        process.spawnerr = reason
        process.give_up()

    # Supervisor issue #591
    def enable_program(self, program_name: str) -> None:
        """ Re-enable the processes to be started and trigger their autostart if configured to.

        :param program_name: the program to enable
        :return: None
        """
        for group_name, group in self.supervisord.process_groups.items():
            for process in group.processes.values():
                if process.supvisors_config.program_config.name == program_name:
                    process.supvisors_config.program_config.disabled = False
                    # fire event to Supervisor listeners
                    notify(ProcessEnabledEvent(process))
                    # if autostart configured, process transition will do the job if laststart is reset
                    process.laststart = 0
                    process.laststart_monotonic = 0.0
                    process.transition()

    def disable_program(self, program_name: str) -> None:
        """ Disable the processes so that they cannot be started.
        It is assumed here that they have been stopped properly before.

        :param program_name: the program to disable
        :return: None
        """
        for group_name, group in self.supervisord.process_groups.items():
            for process in group.processes.values():
                if process.supvisors_config.program_config.name == program_name:
                    process.supvisors_config.program_config.disabled = True
                    # fire event to Supervisor listeners
                    notify(ProcessDisabledEvent(process))

    # Supervisor issue #177
    def add_supervisor_process(self, group_name: str, process_config: ProcessConfig,
                               supvisors_config: SupvisorsProcessConfig) -> Optional[str]:
        """ Add new processes to the Supervisor group from the configuration built.

        :param group_name: the group that embed the program definition.
        :param process_config: the new process configurations to add to the group.
        :param supvisors_config: set to True if the program is disabled.
        :return: the new process namespecs.
        """
        # add new process configs to group in Supervisor
        group = self.supervisord.process_groups[group_name]
        # check if there's a corresponding process in the group (obsolete expected)
        process = group.processes.get(process_config.name)
        if process:
            if not process.obsolete:
                self.logger.error('SupervisorData.add_supervisor_process: unexpected non-obsolete'
                                  f' process={process_config.name} in group={group_name}')
            process.obsolete = False
            return None
        # create processes from new process configs
        self.logger.info(f'SupervisorData.add_supervisor_process: add process={process_config.name}')
        # WARN: replace process_config Supvisors server_options by Supervisor options
        #       this is causing "reaped unknown pid" at exit due to inadequate pidhistory
        process_config.options = self.supervisord.options
        # create a new process from the config
        process = process_config.make_process(group)
        # additional Supvisors attributes
        self.complete_process(process, supvisors_config)
        # prepare log files
        process_config.create_autochildlogs()
        # add the new process to the group processes
        group.processes[process_config.name] = process
        # add the new process config to the group process configs (used by supervisor.getAllConfigInfo)
        group.config.process_configs.append(process_config)
        # fire event to Supervisor listeners
        notify(ProcessAddedEvent(process))
        return make_namespec(group_name, process_config.name)

    def obsolete_processes(self, group_processes: GroupConfigInfo) -> NameList:
        """ Mark the selected processes as obsolete. """
        obsolete_processes = []
        for group_name, candidates in group_processes.items():
            group = self.supervisord.process_groups[group_name]
            for process_name in candidates:
                process = group.processes[process_name]
                if process.state in STOPPED_STATES:
                    # deleted processes are not added to the list of obsolete processes
                    self.delete_process(group, process)
                else:
                    process.obsolete = True
                    # add the process namespec to the list pf obsolete processes
                    obsolete_processes.append(make_namespec(group_name, process_name))
        return obsolete_processes

    @staticmethod
    def complete_process(process: Subprocess, supvisors_config: SupvisorsProcessConfig):
        """ Add addition configuration to the process and to its config structure. """
        # prepare monotonic start/stop times
        process.laststart_monotonic = 0.0
        process.laststop_monotonic = 0.0
        # prepare the process obsolescence (in case of numprocs decrease)
        process.obsolete = False
        # prepare extra arguments
        process.extra_args = ''
        # store the extended process config
        process.supvisors_config = supvisors_config

    def delete_process(self, group: ProcessGroupBase, process: Subprocess):
        """ Delete a process from Supervisor internal data. """
        self.logger.warn(f'SupervisorData.delete_process: delete process={process.config.name}'
                         f' in group={group.config.name}')
        # fire event to Supervisor listeners
        notify(ProcessRemovedEvent(process))
        # delete the process from the group
        del group.processes[process.config.name]
        # delete the process config
        group.config.process_configs.remove(process.config)



# ======================================================================
# Copyright 2020 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import os
from unittest.mock import call, Mock

import pytest
import supervisor

from supvisors.web.viewmaintail import *
from .base import DummyHttpContext


@pytest.fixture
def http_context(supvisors):
    """ Fixture for a consistent mocked HTTP context provided by Supervisor. """
    http_context = DummyHttpContext()
    # use a supervisor template for once
    supervisor_path = next(iter(supervisor.__path__), '.')
    http_context.template = os.path.join(supervisor_path, 'ui/tail.html')
    # assign supvisors in supervisord structure
    http_context.supervisord.supvisors = supvisors
    supvisors.supervisor_data.supervisord = http_context.supervisord
    return http_context


@pytest.fixture
def view(http_context):
    """ Return the instance to test. """
    # create the instance to be tested
    return MainTailView(http_context)


def test_init(view):
    """ Test the construction of MainTailView. """
    # test instance inheritance
    assert isinstance(view, MeldView)


def test_render(mocker, view):
    """ Test the MainTailView.render method. """
    rpc_intf = view.context.supervisord.supvisors.supervisor_data.supervisor_rpc_interface
    rpc = mocker.patch.object(rpc_intf, 'readLog', return_value='supervisor log tail')
    # build XML context
    title_elt = Mock()
    body_elt = Mock()
    anchor_elt = Mock()
    mid_map = {'title': title_elt, 'tailbody': body_elt, 'refresh_anchor': anchor_elt}
    mocked_root = Mock(**{'findmeld.side_effect': lambda x: mid_map[x],
                          'write_xhtmlstring.return_value': 'xhtml'})
    mocker.patch('supervisor.web.MeldView.clone', return_value=mocked_root)
    # test RPC error NO_FILE
    rpc.side_effect = RPCError(Faults.NO_FILE)
    assert view.render() == 'xhtml'
    assert title_elt.content.call_args_list == [call('Supervisor tail')]
    assert body_elt.content.call_args_list == [call('No file for Supervisor')]
    assert anchor_elt.attributes.call_args_list == [call(href='maintail.html?limit=1024')]
    for elt in mid_map.values():
        elt.reset_mock()
    # test RPC error BAD_ARGUMENTS
    rpc.side_effect = RPCError(Faults.BAD_ARGUMENTS, 'invalid limit')
    assert view.render() == 'xhtml'
    assert title_elt.content.call_args_list == [call('Supervisor tail')]
    assert body_elt.content.call_args_list == [call('ERROR: unexpected rpc fault [3] BAD_ARGUMENTS: invalid limit')]
    assert anchor_elt.attributes.call_args_list == [call(href='maintail.html?limit=1024')]
    for elt in mid_map.values():
        elt.reset_mock()
    # test normal case
    rpc.side_effect = None
    view.context.form['limit'] = '2048'
    assert view.render() == 'xhtml'
    assert title_elt.content.call_args_list == [call('Supervisor tail')]
    assert body_elt.content.call_args_list == [call('supervisor log tail')]
    assert anchor_elt.attributes.call_args_list == [call(href='maintail.html?limit=2048')]
